
#   Foremast - Pipeline Tooling
#
#   Copyright 2019 Redbox Automated Retail, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""AWS Auto Scaling Group related utilities."""
import logging

from ..utils.gate import gate_request

LOG = logging.getLogger(__name__)


def get_latest_server_group(env, app):
    """Finds the most recently deployed server group for the application.

    Returns:
        server_group (str): Name of the newest server group
    """
    uri = "/applications/{0}".format(app)
    response = gate_request(uri=uri)

    for server_group in response.json()['clusters'][env]:
        LOG.debug("Server Group Response: %s", server_group)
        return server_group['serverGroups'][-1]



"""Tests for security group duplicates."""
from unittest import mock

from foremast.utils.security_group  import remove_duplicate_sg

SECURITYGROUP_REPLACEMENTS = {
    'enabled_new_sg': 'old_sg'
}

SECURITY_GROUP_DATA = ['old_sg', 'current_sg', 'enabled_new_sg'] 
VALID_SECURITY_GROUPS = ['current_sg', 'enabled_new_sg']


def test_duplicate_sg_removal():
    """Check that duplicate SGs are removed based on SECURITYGROUP_REPLACEMENTS const."""
    with mock.patch.dict('foremast.utils.security_group.SECURITYGROUP_REPLACEMENTS', SECURITYGROUP_REPLACEMENTS):
        security_groups = remove_duplicate_sg(SECURITY_GROUP_DATA)
        assert security_groups == VALID_SECURITY_GROUPS



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Create lambda functions"""

import logging
import zipfile
from os.path import exists

import boto3
from botocore.exceptions import ClientError
from tryagain import retries

from ..consts import LAMBDA_STANDALONE_MODE
from ..exceptions import RequiredKeyNotFound
from ..utils import get_details, get_lambda_arn, get_properties, get_role_arn, get_security_group_id, get_subnets, \
    exponential_backoff

LOG = logging.getLogger(__name__)


class LambdaFunction:
    """Manipulate Lambda function."""

    def __init__(self, app, env, region, prop_path, artifact_path):
        """Lambda function object.

        Args:
            app (str): Application name
            env (str): Environment/Account
            region (str): AWS Region
            prop_path (str): Path of environment property file
            artifact_path (str): Path or URI for code artifact
        """
        self.app_name = app
        self.env = env
        self.region = region
        self.properties = get_properties(prop_path)
        generated = get_details(app=self.app_name)
        self.group = generated.data['project']
        self.artifact_path = artifact_path

        try:
            self.pipeline = self.properties['pipeline']['lambda']
        except KeyError:
            raise RequiredKeyNotFound("Lambda key in pipeline.json is required.")

        self.runtime = self.pipeline['runtime']
        self.description = self.pipeline['app_description']
        self.handler = self.pipeline['handler']
        self.vpc_enabled = self.pipeline['vpc_enabled']
        self.package_type = "zip" if 'package_type' not in self.pipeline else self.pipeline['package_type']

        self.settings = get_properties(prop_path, env=self.env, region=self.region)
        app = self.settings['app']
        self.custom_tags = app['custom_tags']
        self.lambda_environment = app['lambda_environment']
        self.lambda_layers = app['lambda_layers']
        self.lambda_destinations = app['lambda_destinations']
        self.lambda_dlq = app['lambda_dlq']
        self.lambda_filesystems = app['lambda_filesystems']
        self.lambda_tracing = app['lambda_tracing']
        self.lambda_provisioned_throughput = app['lambda_provisioned_throughput']
        self.memory = app['lambda_memory']
        self.role = app.get('lambda_role') or generated.iam()['lambda_role']
        self.timeout = app['lambda_timeout']
        self.concurrency_limit = app.get('lambda_concurrency_limit')
        self.subnet_purpose = app['lambda_subnet_purpose']
        self.subnet_count = app['lambda_subnet_count']

        self.role_arn = get_role_arn(self.role, self.env, self.region)

        self.session = boto3.Session(profile_name=self.env, region_name=self.region)
        self.lambda_client = self.session.client('lambda')

    def _check_lambda(self):
        """Check if lambda function exists.

        Returns:
            True if function does exist
            False if function does not exist
        """
        try:
            self.lambda_client.get_function(FunctionName=self.app_name)
            return True
        except boto3.exceptions.botocore.exceptions.ClientError:
            return False

    def _check_lambda_alias(self):
        """Check if lambda alias exists.

        Returns:
            True if alias exists
            False if alias does not exist
        """
        aliases = self.lambda_client.list_aliases(FunctionName=self.app_name)

        matched_alias = False
        for alias in aliases['Aliases']:
            if alias['Name'] == self.env:
                LOG.info('Found alias %s for function %s', self.env, self.app_name)
                matched_alias = True
                break
        else:
            LOG.info('No alias %s found for function %s', self.env, self.app_name)
        return matched_alias

    def _vpc_config(self):
        """Get VPC config."""
        if self.vpc_enabled:
            subnets_data = get_subnets(env=self.env, region=self.region, purpose=self.subnet_purpose)
            subnets = subnets_data['subnet_ids'][self.region]
            if self.subnet_count:
                subnets = subnets[:self.subnet_count]
                LOG.info('Subnet Count of %s specified. Limiting to subnets: %s', self.subnet_count, subnets)

            security_groups = self._get_sg_ids()

            vpc_config = {'SubnetIds': subnets, 'SecurityGroupIds': security_groups}
        else:
            vpc_config = {'SubnetIds': [], 'SecurityGroupIds': []}
        LOG.debug("Lambda VPC config setup: %s", vpc_config)
        return vpc_config

    def _get_sg_ids(self):
        """Get IDs for all defined security groups.

        Returns:
            list: security group IDs for all lambda_extras
        """
        try:
            lambda_extras = self.settings['security_groups']['lambda_extras']
        except KeyError:
            lambda_extras = []

        security_groups = [self.app_name] + lambda_extras
        sg_ids = []
        for security_group in security_groups:
            sg_id = get_security_group_id(name=security_group, env=self.env, region=self.region)
            sg_ids.append(sg_id)
        return sg_ids

    @retries(max_attempts=3, wait=exponential_backoff, exceptions=ClientError)
    def _create_alias(self):
        """Create lambda alias with env name and points it to $LATEST."""
        LOG.info('Creating alias %s', self.env)

        try:
            self.lambda_client.create_alias(
                FunctionName=self.app_name,
                Name=self.env,
                FunctionVersion='$LATEST',
                Description='Alias for {}'.format(self.env))
        except boto3.exceptions.botocore.exceptions.ClientError as error:
            LOG.debug('Create alias error: %s', error)
            LOG.info("Alias creation failed. Retrying...")
            raise

    @retries(max_attempts=3, wait=exponential_backoff, exceptions=ClientError)
    def _update_alias(self):
        """Update lambda alias to point to $LATEST."""
        LOG.info('Updating alias %s to point to $LATEST', self.env)

        try:
            self.lambda_client.update_alias(FunctionName=self.app_name, Name=self.env, FunctionVersion='$LATEST')
        except boto3.exceptions.botocore.exceptions.ClientError as error:
            LOG.debug('Update alias error: %s', error)
            LOG.info("Alias update failed. Retrying...")
            raise

    @retries(max_attempts=5, wait=lambda n: exponential_backoff(n, 4), exceptions=ClientError)
    def _update_function_code(self):
        code_args = self._get_default_lambda_code()
        self.lambda_client.update_function_code(
            FunctionName=self.app_name,
            **code_args
        )
        LOG.info("Successfully updated Lambda code.")

    @retries(max_attempts=5, wait=lambda n: exponential_backoff(n, 4), exceptions=ClientError)
    def _update_function_configuration(self, vpc_config):
        """Update existing Lambda function configuration.

        Args:
            vpc_config (dict): Dictionary of SubnetIds and SecurityGroupsIds for using
                               a VPC in lambda
        """
        LOG.info('Updating configuration for lambda function: %s', self.app_name)

        default_tags = {'app_group': self.group, 'app_name': self.app_name}
        lambda_tags = {**default_tags, **self.custom_tags}

        try:
            lambda_args = self._get_lambda_args("update", vpc_config, lambda_tags)
            self.lambda_client.update_function_configuration(**lambda_args)

            if self._check_lambda_alias():
                self._update_alias()
            else:
                self._create_alias()

            self._put_concurrent_limits(delete_old_config=True)
            self._put_destinations()
            self._put_provisioned_throughput(delete_old_config=True)

        except boto3.exceptions.botocore.exceptions.ClientError as error:
            if 'CreateNetworkInterface' in error.response['Error']['Message']:
                message = '{0} is missing "ec2:CreateNetworkInterface"'.format(self.role_arn)
                LOG.debug(message)
                raise SystemExit(message)

            LOG.warning("Client error {} during {}.  Retrying with backoff."
                        .format(error.response['Error']['Code'], error.operation_name))
            raise
        LOG.info('Updating Lambda function tags')

        lambda_arn = get_lambda_arn(self.app_name, self.env, self.region)
        self.lambda_client.tag_resource(Resource=lambda_arn, Tags=lambda_tags)

        LOG.info("Successfully updated Lambda configuration.")

    @retries(max_attempts=3, wait=exponential_backoff, exceptions=ClientError)
    def _put_provisioned_throughput(self, delete_old_config):
        """Update existing Lambda function provisioned throughput config

            Args:
                delete_old_config (bool): delete existing config if no provisioned
                throughput set in pipeline files
            """
        try:
            if self.lambda_provisioned_throughput:
                self.lambda_client.put_provisioned_concurrency_config(
                    FuctionName=self.app_name,
                    Qualifier=self.env,
                    ProvisionedConcurrentExecutions=self.lambda_provisioned_throughput)
            elif delete_old_config:
                self.lambda_client.delete_provisioned_concurrency_config(
                    FunctionName=self.app_name,
                    Qualifier=self.env)
        except ClientError as error:
            LOG.warning("Client error {} during {}.  Retrying with backoff."
                        .format(error.response['Error']['Code'], error.operation_name))
            raise

    @retries(max_attempts=3, wait=exponential_backoff, exceptions=ClientError)
    def _put_destinations(self):
        try:
            if self.lambda_destinations:
                self.lambda_client.put_function_event_invoke_config(
                    FunctionName=self.app_name,
                    DestinationConfig=self.lambda_destinations
                )
        except ClientError as error:
            LOG.warning("Client error {} during {}.  Retrying with backoff."
                        .format(error.response['Error']['Code'], error.operation_name))
            raise

    @retries(max_attempts=3, wait=exponential_backoff, exceptions=ClientError)
    def _put_concurrent_limits(self, delete_old_config):
        """Update existing Lambda function concurrency

        Args:
            delete_old_config (bool): delete existing config if no concurrency
            limit set in pipeline files
        """
        try:
            if self.concurrency_limit:
                self.lambda_client.put_function_concurrency(
                    FunctionName=self.app_name,
                    ReservedConcurrentExecutions=self.concurrency_limit
                )
            elif delete_old_config:
                self.lambda_client.delete_function_concurrency(FunctionName=self.app_name)
        except ClientError as error:
            LOG.warning("Client error {} during {}.  Retrying with backoff."
                        .format(error.response['Error']['Code'], error.operation_name))
            raise

    @retries(max_attempts=3, wait=exponential_backoff, exceptions=SystemExit)
    def _create_function(self, vpc_config):
        """Create lambda function, configures lambda parameters.

        We need to upload non-zero zip when creating function. Uploading
        hello_world python lambda function since AWS doesn't care which
        executable is in ZIP.

        Args:
            vpc_config (dict): Dictionary of SubnetIds and SecurityGroupsIds for using
                               a VPC in lambda
        """
        LOG.info('Creating lambda function: %s', self.app_name)

        default_tags = {'app_group': self.group, 'app_name': self.app_name}
        lambda_tags = {**default_tags, **self.custom_tags}

        try:
            lambda_args = self._get_lambda_args("create", vpc_config, lambda_tags)
            self.lambda_client.create_function(**lambda_args)

            if self._check_lambda_alias():
                self._update_alias()
            else:
                self._create_alias()
            
            self._put_concurrent_limits(delete_old_config=False)
            self._put_destinations()
            self._put_provisioned_throughput(delete_old_config=False)

        except ClientError as error:
            if 'CreateNetworkInterface' in error.response['Error']['Message']:
                message = '{0} is missing "ec2:CreateNetworkInterface"'.format(self.role_arn)
                LOG.critical(message)
                raise SystemExit(message)

            raise

        LOG.info("Successfully created Lambda function and alias")

    def deploy_lambda_function(self):
        """Create or update Lambda function."""
        vpc_config = self._vpc_config()

        if self._check_lambda():
            self._update_function_configuration(vpc_config)
            if LAMBDA_STANDALONE_MODE:
                self._update_function_code()
        else:
            self._create_function(vpc_config)

    def _get_lambda_args(self, action, vpc_config, lambda_tags):
        """Gets lambda args as a dictionary, depending on properties such as package_type.
        Args:
            action (str): Action taken place, either create or update.  This adjust the
                          outputted dictionary and it's keys
        """
        # Default args that apply to all package types
        lambda_args = {
            "Environment": self.lambda_environment,
            "FunctionName": self.app_name,
            "Role": self.role_arn,
            "Description":  self.description,
            "Timeout": int(self.timeout),
            "MemorySize": int(self.memory),
            "VpcConfig": vpc_config,
            "Layers": self.lambda_layers,
            "DeadLetterConfig": self.lambda_dlq,
            "TracingConfig": self.lambda_tracing,
            "FileSystemConfigs": self.lambda_filesystems,
        }

        if action == "create":
            lambda_args["Publish"] = True if LAMBDA_STANDALONE_MODE else False
            lambda_args["PackageType"] = self.package_type.capitalize()
            lambda_args["Tags"] = lambda_tags
            lambda_args["Code"] = self._get_default_lambda_code()

        # Args for zip packages only
        if self.package_type.lower() == "zip":
            lambda_args["Runtime"] = self.runtime
            lambda_args["Handler"] = self.handler

        return lambda_args

    def _get_default_lambda_code(self) -> dict:
        if self.package_type.lower() == "zip":
            return self._get_default_lambda_code_zip()
        elif self.package_type.lower() == "image":
            return {'ImageUri': self.artifact_path}
        else:
            raise Exception("Invalid Lambda package_type: " + self.package_type)

    def _get_default_lambda_code_zip(self) -> dict:
        if LAMBDA_STANDALONE_MODE:
            if self.artifact_path and exists(self.artifact_path):
                with open(self.artifact_path, 'rb') as openfile:
                    contents = openfile.read()
                    return {'ZipFile': contents}
            raise Exception("Artifact file '{}' not found.  Note: You must download and pass in a local address"
                            .format(self.artifact_path))
        else:
            return {'ZipFile': self._create_dummy_zip_package()}

    def _create_dummy_zip_package(self) -> bytes:
        """When lambda's are created during foremast infra step, we must create a dummy zip as a placeholder
        until the real code is added in next Spinnaker step.  This does not apply if using feature flag
        LAMBDA_STANDALONE_MODE=True."""
        LOG.info("Creating dummy zip package and overriding artifact path")
        self.artifact_path = 'lambda-holder.zip'
        with zipfile.ZipFile(self.artifact_path, mode='w') as zipped:
            zipped.writestr('index.py', 'print "Hello world"')
        with open('lambda-holder.zip', 'rb') as openfile:
            return openfile.read()



"""Test S3 tags result."""
from foremast.utils import generate_s3_tags


def test_s3_tags_data():
    """Result data should be list then needs to have Key:key Value:value format."""
    test_result = generate_s3_tags.generated_tag_data({'app_group': 'test_value'})

    assert list == type(test_result)
    assert 'app_group' == test_result[0]['Key']
    assert 'test_value' == test_result[0]['Value']



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Test ELB creation functions."""
import json
from unittest import mock

from foremast.elb import SpinnakerELB
from foremast.elb.format_listeners import format_cert_name, format_listeners
from foremast.elb.splay_health import splay_health

SAMPLE_TLSCERT_V1_JSON = """
{
  "dev": {
    "wildcard.example.com": "arn:aws:iam::123456789012:server-certificate/wildcard.example.com-2020-07-15",
    "wildcard.dev.example.com": "arn:aws:acm:us-east-1:123456789012:certificate/11111111-2222-3333-4444-555555555555"
  },
  "prod": {
    "wildcard.example.com": "arn:aws:iam::210987654321:server-certificate/wildcard.example.com-2020-07-15",
    "wildcard.prod.example.com": "arn:aws:acm:us-east-1:210987654321:certificate/11111111-2222-3333-4444-555555555555",
    "wildcard.us-west-2.prod.example.com": "arn:aws:acm:us-west-2:210987654321:certificate/11111111-0000-2222-3333-555555555555"
  }
}
"""

SAMPLE_TLSCERT_V2_JSON = """{
  "iam": {
    "dev": {
      "wildcard.example.com": "arn:aws:iam::123456789012:server-certificate/wildcard.example.com-2020-07-15"
    },
    "prod": {
      "wildcard.example.com": "arn:aws:iam::210987654321:server-certificate/wildcard.example.com-2020-07-15"
    }
  },
  "acm": {
    "us-east-1": {
      "dev": {
        "wildcard.dev.example.com": "arn:aws:acm:us-east-1:123456789012:certificate/11111111-2222-3333-4444-555555555555"
      },
      "prod": {
        "wildcard.prod.example.com": "arn:aws:acm:us-east-1:210987654321:certificate/11111111-2222-3333-4444-555555555555"
      }
    },
    "us-west-2": {
      "dev": {
        "wildcard.dev.example.com": "arn:aws:acm:us-west-2:123456789012:certificate/11111111-0000-2222-3333-555555555555"
      },
      "prod": {
        "wildcard.prod.example.com": "arn:aws:acm:us-west-2:210987654321:certificate/11111111-0000-2222-3333-555555555555"
      }
    }
  }
}"""

def test_elb_splay():
    """Splay should split Health Checks properly."""
    health = splay_health('HTTP:80/test')
    assert health.path == '/test'
    assert health.port == '80'
    assert health.proto == 'HTTP'
    assert health.target == 'HTTP:80/test'

    health = splay_health('TCP:8000/test')
    assert health.path == ''
    assert health.port == '8000'
    assert health.proto == 'TCP'
    assert health.target == 'TCP:8000'

    health = splay_health('HTTPS:8000/test')
    assert health.path == '/test'
    assert health.port == '8000'
    assert health.proto == 'HTTPS'
    assert health.target == 'HTTPS:8000/test'

    health = splay_health('HTTPS:80')
    assert health.path == '/healthcheck'
    assert health.port == '80'
    assert health.proto == 'HTTPS'
    assert health.target == 'HTTPS:80/healthcheck'


@mock.patch('foremast.elb.format_listeners.get_env_credential')
def test_elb_format_listeners(mock_creds):
    """Listeners should be formatted in list of dicts."""
    mock_creds.return_value = {'accountId': '0100'}

    config = {
        'certificate': None,
        'i_port': 8080,
        'i_proto': 'HTTP',
        'lb_port': 80,
        'lb_proto': 'HTTP',
        'policies': [],
        'listener_policies': [],
        'backend_policies': [],
    }
    generated = [{
        'externalPort': 80,
        'externalProtocol': 'HTTP',
        'internalPort': 8080,
        'internalProtocol': 'HTTP',
        'sslCertificateId': None,
        'listenerPolicies': [],
        'backendPolicies': [],
    }]

    # check defaults
    assert generated == format_listeners(elb_settings=config)

    # 'ports' key should override old style definitions
    config['ports'] = [{'instance': 'HTTP:8000', 'loadbalancer': 'http:500'}]
    generated = [{
        'externalPort': 500,
        'externalProtocol': 'HTTP',
        'internalPort': 8000,
        'internalProtocol': 'HTTP',
        'sslCertificateId': None,
        'listenerPolicies': [],
        'backendPolicies': [],
    }]

    # check ports
    assert generated == format_listeners(elb_settings=config)

    config['ports'].append({
        'certificate': 'kerby',
        'instance': 'http:80',
        'loadbalancer': 'https:443',
    })
    generated.append({
        'externalPort': 443,
        'externalProtocol': 'HTTPS',
        'internalPort': 80,
        'internalProtocol': 'HTTP',
        'sslCertificateId': 'arn:aws:iam::0100:server-certificate/kerby',
        'listenerPolicies': [],
        'backendPolicies': [],
    })

    # check certificate
    assert generated == format_listeners(elb_settings=config)


def test_elb_format_cert_name():
    """Test the format_cert_name method"""
    assert None == format_cert_name()

    full_cert_name = 'arn:aws:123'
    assert full_cert_name == format_cert_name(certificate=full_cert_name)

    compiled_cert = 'arn:aws:iam::123456789012:server-certificate/mycert1'
    assert compiled_cert == format_cert_name(account='123456789012', certificate='mycert1')

@mock.patch("foremast.elb.format_listeners.get_template")
def test_elb_cert_name_v1(rendered_template):
    """Tests the format_cert_name method when used with v1 template"""
    rendered_template.return_value = SAMPLE_TLSCERT_V1_JSON

    iam_cert = "arn:aws:iam::210987654321:server-certificate/wildcard.example.com-2020-07-15"
    assert iam_cert == format_cert_name(env='prod', account='210987654321', region='us-east-1', certificate='wildcard.example.com')

    acm_cert = 'arn:aws:acm:us-east-1:210987654321:certificate/11111111-2222-3333-4444-555555555555'
    assert acm_cert == format_cert_name(env='prod', account='210987654321', region='us-east-1', certificate='wildcard.prod.example.com')

    acm_region_cert = 'arn:aws:acm:us-west-2:210987654321:certificate/11111111-0000-2222-3333-555555555555'
    assert acm_region_cert == format_cert_name(env='prod', account='210987654321', region='us-west-2', certificate='wildcard.us-west-2.prod.example.com')


@mock.patch("foremast.elb.format_listeners.get_template")
def test_elb_cert_name_v2(rendered_template):
    """Tests the format_cert_name method when used with v2 template"""
    rendered_template.return_value = SAMPLE_TLSCERT_V2_JSON

    iam_cert = "arn:aws:iam::210987654321:server-certificate/wildcard.example.com-2020-07-15"
    assert iam_cert == format_cert_name(env='prod', account='210987654321', region='us-east-1', certificate='wildcard.example.com')

    acm_cert = 'arn:aws:acm:us-east-1:210987654321:certificate/11111111-2222-3333-4444-555555555555'
    assert acm_cert == format_cert_name(env='prod', account='210987654321', region='us-east-1', certificate='wildcard.prod.example.com')

    acm_region_cert = 'arn:aws:acm:us-west-2:210987654321:certificate/11111111-0000-2222-3333-555555555555'
    assert acm_region_cert == format_cert_name(env='prod', account='210987654321', region='us-west-2', certificate='wildcard.prod.example.com')


@mock.patch.object(SpinnakerELB, 'configure_attributes')
@mock.patch.object(SpinnakerELB, 'add_backend_policy')
@mock.patch.object(SpinnakerELB, 'add_listener_policy')
@mock.patch('foremast.elb.create_elb.wait_for_task')
@mock.patch.object(SpinnakerELB, 'make_elb_json', return_value={})
@mock.patch('foremast.elb.create_elb.get_properties')
def test_elb_create_elb(mock_get_properties, mock_elb_json, mock_wait_for_task, mock_listener_policy,
                        mock_backend_policy, mock_load_balancer_attributes):
    """Test SpinnakerELB create_elb method"""
    elb = SpinnakerELB(app='myapp', env='dev', region='us-east-1')
    elb.create_elb()
    mock_listener_policy.assert_called_with(mock_elb_json())


@mock.patch.dict('foremast.elb.create_elb.DEFAULT_ELB_SECURITYGROUPS', {"dev": []})
@mock.patch('foremast.elb.create_elb.get_vpc_id', return_value='vpc-100')
@mock.patch('foremast.elb.create_elb.format_listeners')
@mock.patch('foremast.elb.create_elb.get_subnets')
@mock.patch('foremast.elb.create_elb.get_properties')
def test_elb_make_elb_json(mock_get_properties, mock_get_subnets, mock_format_listeners, mock_vpc_id):
    """Test SpinnakerELB make_elbj_json method"""
    properties = {
        'elb': {
            'health': {
                'interval': 10,
                'timeout': 1,
                'threshold': 2,
                'unhealthy_threshold': 3,
            }
        },
        'security_group': {
            'elb_extras': []
        },
    }

    subnets = {'us-east-1': ['subnet-1']}
    listeners = []

    mock_get_properties.return_value = properties
    mock_get_subnets.return_value = subnets
    mock_format_listeners.return_value = listeners

    # internal elb
    elb = SpinnakerELB(app='myapp', env='dev', region='us-east-1')
    elb_str = elb.make_elb_json()
    assert isinstance(elb_str, str)
    elb_json = json.loads(elb_str)
    assert elb_json['job'][0]['isInternal']

    # external elb
    properties['elb'].update({'subnet_purpose': 'external'})
    elb = SpinnakerELB(app='myapp', env='dev', region='us-east-1')
    elb_json = json.loads(elb.make_elb_json())
    assert not elb_json['job'][0]['isInternal']


@mock.patch('foremast.elb.create_elb.boto3.session.Session')
@mock.patch('foremast.elb.create_elb.get_properties')
def test_elb_add_listener_policy(mock_get_properties, mock_boto3_session):
    test_app = 'myapp'
    test_port = 80
    test_policy_list = ['policy_name']

    json_data = {
        'job': [{
            'listeners': [
                {
                    'externalPort': test_port,
                    'listenerPolicies': test_policy_list
                },
            ],
        }],
    }
    client = mock_boto3_session.return_value.client.return_value

    elb = SpinnakerELB(app='myapp', env='dev', region='us-east-1')
    elb.add_listener_policy(json.dumps(json_data))
    client.set_load_balancer_policies_of_listener.assert_called_with(
        LoadBalancerName=test_app, LoadBalancerPort=test_port, PolicyNames=test_policy_list)


@mock.patch('foremast.elb.create_elb.boto3.session.Session')
@mock.patch('foremast.elb.create_elb.get_properties')
def test_elb_add_backend_policy(mock_get_properties, mock_boto3_session):
    test_app = 'myapp'
    test_port = 80
    test_policy_list = ['policy_name']

    json_data = {
        'job': [{
            'listeners': [
                {
                    'internalPort': test_port,
                    'backendPolicies': test_policy_list
                },
            ],
        }],
    }
    client = mock_boto3_session.return_value.client.return_value

    elb = SpinnakerELB(app='myapp', env='dev', region='us-east-1')
    elb.add_backend_policy(json.dumps(json_data))
    client.set_load_balancer_policies_for_backend_server.assert_called_with(
        LoadBalancerName=test_app, InstancePort=test_port, PolicyNames=test_policy_list)



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Get available Subnets for specific Targets."""
import logging
from collections import defaultdict
from pprint import pformat

from tryagain import retries

from ..exceptions import SpinnakerSubnetError, SpinnakerTimeout
from ..utils.gate import gate_request

LOG = logging.getLogger(__name__)


# TODO: split up into get_az, and get_subnet_id
@retries(max_attempts=6, wait=2.0, exceptions=SpinnakerTimeout)  # noqa
def get_subnets(target='ec2', purpose='internal', env='', region=''): # noqa
    """Get all availability zones for a given target.

    Args:
        target (str): Type of subnets to look up (ec2 or elb).
        env (str): Environment to look up.
        region (str): AWS Region to find Subnets for.

    Returns:
        az_dict: dictionary of  availbility zones, structured like
        { $region: [ $avaibilityzones ] }
        or
        { $account: $region: [ $availabilityzone] }
    """
    account_az_dict = defaultdict(defaultdict)
    subnet_id_dict = defaultdict(defaultdict)

    uri = '/subnets/aws'
    subnet_response = gate_request(uri=uri)

    if not subnet_response.ok:
        raise SpinnakerTimeout(subnet_response.text)

    subnet_list = subnet_response.json()
    for subnet in subnet_list:
        LOG.debug('Subnet Response: %s', subnet)

        if subnet.get('target', '') == target:
            availability_zone = subnet['availabilityZone']
            account = subnet['account']
            subnet_region = subnet['region']
            subnet_id = subnet['id']
            try:
                if availability_zone not in account_az_dict[account][subnet_region]:
                    account_az_dict[account][subnet_region].append(availability_zone)
            except KeyError:
                account_az_dict[account][subnet_region] = [availability_zone]
            # get list of all subnet IDs with correct purpose
            if subnet['purpose'] == purpose:
                try:
                    subnet_id_dict[account][subnet_region].append(subnet_id)
                except KeyError:
                    subnet_id_dict[account][subnet_region] = [subnet_id]

            LOG.debug('%s regions: %s', account, list(account_az_dict[account].keys()))

    if all([env, region]):
        try:
            region_dict = {region: account_az_dict[env][region]}
            region_dict['subnet_ids'] = {region: subnet_id_dict[env][region]}
            LOG.debug('Region dict: %s', region_dict)
            return region_dict
        except KeyError:
            raise SpinnakerSubnetError(env=env, region=region)

    LOG.debug('AZ dict:\n%s', pformat(dict(account_az_dict)))

    return account_az_dict



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Create onetime Pipelines for Spinnaker.

These are circumventions for redployments to a specific Environment in a Region.
"""
import json

from .create_pipeline import SpinnakerPipeline


class SpinnakerPipelineOnetime(SpinnakerPipeline):
    """Manipulate Spinnaker Pipelines.

    Args:
        app (str): Application name.
        trigger_job (str): Jenkins trigger job.
        base (str): Base image name (i.e: fedora).
        prop_path (str): Path to the raw.properties.json.
        onetime (str): Environment to build onetime pipeline for.
    """

    def __init__(self, app='', trigger_job='', prop_path='', base='', onetime='', runway_dir=''):
        super().__init__(
            app=app,
            trigger_job=trigger_job,
            prop_path=prop_path,
            base=base,
            runway_dir=runway_dir, )
        self.environments = [onetime]

    def post_pipeline(self, pipeline):
        """Send Pipeline JSON to Spinnaker.

        Args:
            pipeline (dict, str): New Pipeline to create.
        """
        if isinstance(pipeline, str):
            pipeline_str = pipeline
        else:
            pipeline_str = json.dumps(pipeline)

        pipeline_json = json.loads(pipeline_str)

        # Note pipeline name is manual
        name = '{0} (onetime-{1})'.format(pipeline_json['name'], self.environments[0])
        pipeline_json['name'] = name

        # Inject pipeline Id so that it does not override existing pipelines
        pipeline_id = super().compare_with_existing(onetime=True)
        if pipeline_id:
            pipeline_json['id'] = pipeline_id
        else:
            del pipeline_json['id']

        # disable trigger as not to accidently kick off multiple deployments
        for trigger in pipeline_json['triggers']:
            trigger['enabled'] = False

        self.log.debug('Manual Pipeline JSON:\n%s', pipeline_json)
        super().post_pipeline(pipeline_json)



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""CLI entry point to create Spinnaker Pipelines.

Help: ``python -m src.foremast.pipeline -h``
"""
import argparse
import logging

from ..args import add_app, add_debug, add_properties
from ..consts import ENVS, LOGGING_FORMAT
from .create_pipeline import SpinnakerPipeline
from .create_pipeline_lambda import SpinnakerPipelineLambda
from .create_pipeline_cloudfunction import SpinnakerPipelineCloudFunction
from .create_pipeline_onetime import SpinnakerPipelineOnetime
from .create_pipeline_s3 import SpinnakerPipelineS3


def main():
    """Creates a pipeline in Spinnaker"""
    logging.basicConfig(format=LOGGING_FORMAT)
    log = logging.getLogger(__name__)

    parser = argparse.ArgumentParser()
    add_debug(parser)
    add_app(parser)
    add_properties(parser)
    parser.add_argument('-b', '--base', help='Base AMI name to use, e.g. fedora, tomcat')
    parser.add_argument("--triggerjob", help="The jenkins job to monitor for pipeline triggering", required=True)
    parser.add_argument('--onetime', required=False, choices=ENVS, help='Onetime deployment environment')
    parser.add_argument(
        '-t', '--type', dest='type', required=False, default='ec2', help='Deployment type, e.g. ec2, lambda')
    args = parser.parse_args()

    if args.base and '"' in args.base:
        args.base = args.base.strip('"')

    logging.getLogger(__package__.split('.')[0]).setLevel(args.debug)

    log.debug('Parsed arguments: %s', args)

    if args.onetime:
        spinnakerapps = SpinnakerPipelineOnetime(
            app=args.app, onetime=args.onetime, trigger_job=args.triggerjob, prop_path=args.properties, base=args.base)
        spinnakerapps.create_pipeline()
    else:

        if args.type == "ec2":
            spinnakerapps = SpinnakerPipeline(
                app=args.app, trigger_job=args.triggerjob, prop_path=args.properties, base=args.base)
            spinnakerapps.create_pipeline()
        elif args.type == "lambda":
            spinnakerapps = SpinnakerPipelineLambda(
                app=args.app, trigger_job=args.triggerjob, prop_path=args.properties, base=args.base)
            spinnakerapps.create_pipeline()
        elif args.type == "s3":
            spinnakerapps = SpinnakerPipelineS3(
                app=args.app, trigger_job=args.triggerjob, prop_path=args.properties, base=args.base)
            spinnakerapps.create_pipeline()
        elif args.type == "cloudfunction":
            spinnakerapps = SpinnakerPipelineCloudFunction(
                app=args.app, trigger_job=args.triggerjob, prop_path=args.properties, base=args.base)
            spinnakerapps.create_pipeline()


if __name__ == "__main__":
    main()



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""CLI entry point for creating a Spinnaker application.

Help: ``python -m src.foremast.app -h``
"""
import argparse
import logging

import foremastutils

from ..args import add_app, add_debug, add_properties, add_provider
from ..consts import APP_FORMATS, LOGGING_FORMAT
from ..app import SpinnakerApp
from ..utils import get_properties


def main():
    """Entry point for creating a Spinnaker application."""
    # Setup parser
    parser = argparse.ArgumentParser()
    add_debug(parser)
    add_app(parser)
    add_properties(parser)
    add_provider(parser)
    parser.add_argument(
        '--email', help='Email address to associate with application', default='PS-DevOpsTooling@example.com')
    parser.add_argument('--project', help='Git project to associate with application', default='None')
    parser.add_argument('--repo', help='Git repo to associate with application', default='None')
    parser.add_argument('--git', help='Git URI', default=None)

    args = parser.parse_args()

    logging.basicConfig(format=LOGGING_FORMAT)
    logging.getLogger(__package__.split('.')[0]).setLevel(args.debug)

    if args.git and args.git != 'None':
        parsed = foremastutils.Parser(args.git).parse_url()
        generated = foremastutils.Generator(*parsed, formats=APP_FORMATS)
        project = generated.project
        repo = generated.repo
    else:
        project = args.project
        repo = args.repo

    app_properties = get_properties(args.properties, 'pipeline')
    spinnakerapp = SpinnakerApp(
        provider=args.provider,
        app=args.app,
        email=args.email,
        project=project,
        repo=repo,
        pipeline_config=app_properties
    )
    spinnakerapp.create()


if __name__ == '__main__':
    main()



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Get VPC ID."""
import logging

from ..consts import VPC_NAME
from ..exceptions import SpinnakerVPCIDNotFound, SpinnakerVPCNotFound
from ..utils.gate import gate_request

LOG = logging.getLogger(__name__)


def get_vpc_id(account, region):
    """Get VPC ID configured for ``account`` in ``region``.

    Args:
        account (str): AWS account name.
        region (str): Region name, e.g. us-east-1.

    Returns:
        str: VPC ID for the requested ``account`` in ``region``.

    Raises:
        :obj:`foremast.exceptions.SpinnakerVPCIDNotFound`: VPC ID not found for
            ``account`` in ``region``.
        :obj:`foremast.exceptions.SpinnakerVPCNotFound`: Spinnaker has no VPCs
            configured.

    """
    uri = '/networks/aws'
    response = gate_request(uri=uri)

    if not response.ok:
        raise SpinnakerVPCNotFound(response.text)

    vpcs = response.json()

    for vpc in vpcs:
        LOG.debug('VPC Response: %s', vpc)
        if 'name' in vpc and all([vpc['name'] == VPC_NAME, vpc['account'] == account, vpc['region'] == region]):
            LOG.info('Found VPC ID for %s in %s: %s', account, region, vpc['id'])
            vpc_id = vpc['id']
            break
    else:
        LOG.fatal('VPC list: %s', vpcs)
        raise SpinnakerVPCIDNotFound('No VPC available for {0} [{1}].'.format(account, region))

    return vpc_id



"""Tests for string formatting functions."""
from foremast.utils.pipelines import normalize_pipeline_name

SAFE_PIPELINE_NAMES = {
    '#pudding/pop': '_pudding_pop',
    'caesar?salad': 'caesar_salad',
    'chutes&ladders.json': 'chutes&ladders.json',
    'conga-line': 'conga-line',
    'DAKOTA\\ACCESS': 'DAKOTA_ACCESS',
    'jelly/baby': 'jelly_baby',
    'pizza%2Fpie': 'pizza_2Fpie',
    'water': 'water',
}


def test_safe_pipeline_names():
    """Check Pipeline name safe formatting."""
    for raw, safe in SAFE_PIPELINE_NAMES.items():
        assert normalize_pipeline_name(name=raw) == safe



#   Foremast - Pipeline Tooling
#
#   Copyright 2020 Redbox Automated Retail, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Package for managing IAM in GCP. Spinnaker does not
directly interface with IAM so this package mostly uses Google's cloud packages
"""

import logging
from .policy import get_policy, set_policy, modify_policy_remove_member, modify_policy_add_binding
from .create_iam_resources import GcpIamResourceClient

LOG = logging.getLogger(__name__)



#!/usr/bin/env python
#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""A runner for all of the spinnaker pipe modules.

Read environment variables from Jenkins:

    * EMAIL
    * ENV
    * GIT_REPO
    * PROJECT
    * REGION

Then run specific prepare jobs.
"""
import argparse
import logging
import os

import foremastutils

from . import (autoscaling_policy, awslambda, configs, consts, datapipeline, dns, elb, iam, pipeline, s3,
               scheduled_actions, securitygroup, slacknotify, stepfunction, utils)
from .gcp_iam import GcpIamResourceClient
from .app import SpinnakerApp
from .args import add_debug
from .cloudfunction.cloud_functions_client import CloudFunctionsClient
from .exceptions import ForemastError
from .utils.gcp_environment import GcpEnvironment
from tabulate import tabulate

LOG = logging.getLogger(__name__)


class ForemastRunner:
    """Wrap each pipes module in a way that is easy to invoke."""

    def __init__(self):
        """Setup the Runner for all Foremast modules."""
        debug_flag()

        self.email = os.getenv("EMAIL")
        self.env = os.getenv("ENV")
        self.group = os.getenv("PROJECT")
        self.region = os.getenv("REGION")
        self.repo = os.getenv("GIT_REPO")
        self.runway_dir = os.getenv("RUNWAY_DIR")
        self.artifact_path = os.getenv("ARTIFACT_PATH")
        self.artifact_version = os.getenv("ARTIFACT_VERSION")
        self.artifact_branch = os.getenv("ARTIFACT_BRANCH", "master")
        self.promote_stage = os.getenv("PROMOTE_STAGE", "latest")
        self.provider = os.getenv("PROVIDER", "aws")

        self.git_project = "{}/{}".format(self.group, self.repo)
        parsed = foremastutils.Parser(self.git_project)
        generated = foremastutils.Generator(*parsed.parse_url(), formats=consts.APP_FORMATS)

        self.app = generated.app_name()
        self.trigger_job = generated.jenkins()['name']
        self.git_short = generated.gitlab()['main']

        self.raw_path = "./raw.properties"
        self.json_path = self.raw_path + ".json"
        self.configs = None

    def write_configs(self):
        """Generate the configurations needed for pipes."""
        utils.banner("Generating Configs")
        if not self.runway_dir:
            app_configs = configs.process_git_configs(git_short=self.git_short)
        else:
            app_configs = configs.process_runway_configs(runway_dir=self.runway_dir)

        self.configs = configs.write_variables(
            app_configs=app_configs, out_file=self.raw_path, git_short=self.git_short)

    def create_app(self):
        """Create the spinnaker application."""
        utils.banner("Creating Spinnaker App")
        spinnakerapp = SpinnakerApp(
            provider=self.provider,
            app=self.app,
            email=self.email,
            project=self.group,
            repo=self.repo,
            pipeline_config=self.configs['pipeline'])
        spinnakerapp.create()

    def create_pipeline(self, onetime=None):
        """Create the spinnaker pipeline(s)."""
        utils.banner("Creating Pipeline")

        kwargs = {
            'app': self.app,
            'trigger_job': self.trigger_job,
            'prop_path': self.json_path,
            'base': None,
            'runway_dir': self.runway_dir,
        }

        pipeline_type = self.configs['pipeline']['type']

        if pipeline_type not in consts.ALLOWED_TYPES and pipeline_type not in consts.MANUAL_TYPES:
            raise NotImplementedError('Pipeline type "{0}" not permitted.'.format(pipeline_type))

        if not onetime:
            if pipeline_type == 'lambda':
                spinnakerpipeline = pipeline.SpinnakerPipelineLambda(**kwargs)
            elif pipeline_type == 's3':
                spinnakerpipeline = pipeline.SpinnakerPipelineS3(**kwargs)
            elif pipeline_type == 'datapipeline':
                spinnakerpipeline = pipeline.SpinnakerPipelineDataPipeline(**kwargs)
            elif pipeline_type == 'stepfunction':
                spinnakerpipeline = pipeline.SpinnakerPipelineStepFunction(**kwargs)
            elif pipeline_type in consts.MANUAL_TYPES:
                spinnakerpipeline = pipeline.SpinnakerPipelineManual(**kwargs)
            elif pipeline_type == 'cloudfunction':
                spinnakerpipeline = pipeline.SpinnakerPipelineCloudFunction(**kwargs)
            else:
                # Handles all other pipelines
                spinnakerpipeline = pipeline.SpinnakerPipeline(**kwargs)
        else:
            spinnakerpipeline = pipeline.SpinnakerPipelineOnetime(onetime=onetime, **kwargs)

        spinnakerpipeline.create_pipeline()

    def create_aws_iam(self):
        """Create AWS IAM resources."""
        utils.banner("Creating AWS IAM")
        iam.create_iam_resources(env=self.env, app=self.app)

    def create_gcp_iam(self):
        """Create GCP IAM resources."""
        utils.banner("Creating GCP IAM")
        env = self.get_current_gcp_env()
        gcp_iam_client = GcpIamResourceClient(env=env, app_name=self.app, group_name=self.group,
                                              repo_name=self.repo, configs=self.configs)
        gcp_iam_client.create_iam_resources()

    def create_archaius(self):
        """Create S3 bucket for Archaius."""
        utils.banner("Creating S3")
        s3.init_properties(env=self.env, app=self.app)

    def create_s3app(self):
        """Create S3 infra for s3 applications"""
        utils.banner("Creating S3 App Infrastructure")
        primary_region = self.configs['pipeline']['primary_region']
        s3obj = s3.S3Apps(app=self.app,
                          env=self.env,
                          region=self.region,
                          prop_path=self.json_path,
                          primary_region=primary_region)
        s3obj.create_bucket()

    def deploy_s3app(self):
        """Deploys artifacts contents to S3 bucket"""
        utils.banner("Deploying S3 App")
        primary_region = self.configs['pipeline']['primary_region']
        s3obj = s3.S3Deployment(
            app=self.app,
            env=self.env,
            region=self.region,
            prop_path=self.json_path,
            artifact_path=self.artifact_path,
            artifact_version=self.artifact_version,
            artifact_branch=self.artifact_branch,
            primary_region=primary_region)
        s3obj.upload_artifacts()

    def promote_s3app(self):
        """promotes S3 deployment to LATEST"""
        utils.banner("Promoting S3 App")
        primary_region = self.configs['pipeline']['primary_region']
        s3obj = s3.S3Deployment(
            app=self.app,
            env=self.env,
            region=self.region,
            prop_path=self.json_path,
            artifact_path=self.artifact_path,
            artifact_version=self.artifact_version,
            artifact_branch=self.artifact_branch,
            primary_region=primary_region)
        s3obj.promote_artifacts(promote_stage=self.promote_stage)

    def create_secgroups(self):
        """Create security groups as defined in the configs."""
        utils.banner("Creating Security Group")
        sgobj = securitygroup.SpinnakerSecurityGroup(
            app=self.app, env=self.env, region=self.region, prop_path=self.json_path)
        sgobj.create_security_group()

    def create_awslambda(self):
        """Create lambda, this function is used as apart of Foremast infra command."""
        utils.banner("Creating Lambda Function")
        awslambdaobj = awslambda.LambdaFunction(
            app=self.app, env=self.env, region=self.region, prop_path=self.json_path, artifact_path=None)
        awslambdaobj.deploy_lambda_function()

        utils.banner("Creating Lambda Event")
        lambdaeventobj = awslambda.LambdaEvent(app=self.app, env=self.env, region=self.region, prop_path=self.json_path)
        lambdaeventobj.create_lambda_events()

    def deploy_awslambda(self):
        """Deploys an AWS Lambda, this function is used as apart of the Foremast lambdas deploy command."""
        utils.banner("Deploying Lambda Function")
        lambda_function = awslambda.LambdaFunction(
            app=self.app, env=self.env, region=self.region, prop_path=self.json_path, artifact_path=self.artifact_path)
        lambda_function.deploy_lambda_function()

        utils.banner("Creating Lambda Event")
        lambdaeventobj = awslambda.LambdaEvent(app=self.app, env=self.env, region=self.region, prop_path=self.json_path)
        lambdaeventobj.create_lambda_events()

    def create_elb(self):
        """Create the ELB for the defined environment."""
        utils.banner("Creating ELB")
        elbobj = elb.SpinnakerELB(app=self.app, env=self.env, region=self.region, prop_path=self.json_path)
        elbobj.create_elb()

    def create_dns(self):
        """Create DNS for the defined app and environment."""
        utils.banner("Creating DNS")
        elb_subnet = self.configs[self.env]['elb']['subnet_purpose']
        regions = self.configs[self.env]['regions']
        failover = self.configs[self.env]['dns']['failover_dns']
        primary_region = self.configs['pipeline']['primary_region']
        regionspecific_dns = self.configs[self.env]['dns']['region_specific']

        dnsobj = dns.SpinnakerDns(
            app=self.app, env=self.env, region=self.region, prop_path=self.json_path, elb_subnet=elb_subnet)

        if len(regions) > 1 and failover:
            dnsobj.create_elb_dns(regionspecific=True)
            dnsobj.create_failover_dns(primary_region=primary_region)
        else:
            if regionspecific_dns:
                dnsobj.create_elb_dns(regionspecific=True)

            if self.region == primary_region:
                dnsobj.create_elb_dns(regionspecific=False)

    def create_autoscaling_policy(self):
        """Create Scaling Policy for app in environment"""
        utils.banner("Creating Scaling Policy")
        policyobj = autoscaling_policy.AutoScalingPolicy(
            app=self.app, env=self.env, region=self.region, prop_path=self.json_path)
        policyobj.create_policy()

    def create_scheduled_actions(self):
        """Create Scheduled Actions for app in environment"""
        utils.banner("Creating Scheduled Actions")
        actionsobj = scheduled_actions.ScheduledActions(
            app=self.app, env=self.env, region=self.region, prop_path=self.json_path)
        actionsobj.create_scheduled_actions()

    def create_datapipeline(self):
        """Creates data pipeline and adds definition"""
        utils.banner("Creating Data Pipeline")
        dpobj = datapipeline.AWSDataPipeline(app=self.app, env=self.env, region=self.region, prop_path=self.json_path)
        dpobj.create_datapipeline()
        dpobj.set_pipeline_definition()
        if self.configs[self.env].get('datapipeline').get('activate_on_deploy'):
            dpobj.activate_pipeline()

    def create_stepfunction(self):
        """Creates AWS Step Function"""
        utils.banner("Creating AWS Step Function")
        sfnobj = stepfunction.AWSStepFunction(app=self.app, env=self.env, region=self.region, prop_path=self.json_path)
        sfnobj.create_stepfunction()

    def deploy_cloudfunction(self):
        """Creates a Cloud Function"""
        utils.banner("Creating GCP Cloud Function")
        env = self.get_current_gcp_env()
        cloud_function_client = CloudFunctionsClient(self.app, env, self.configs)
        cloud_function_client.prepare_client()
        cloud_function_client.deploy_function(self.artifact_path, self.region)
        LOG.info("Finished deploying cloud function")

    def slack_notify(self):
        """Send out a slack notification."""
        utils.banner("Sending slack notification")

        if self.env.startswith("prod"):
            notify = slacknotify.SlackNotification(app=self.app, env=self.env, prop_path=self.json_path)
            notify.post_message()
        else:
            LOG.info("No slack message sent, not production environment")

    def get_current_gcp_env(self):
        """Gets the current GCP Environment
        Returns:
            GcpEnvironment, the current GCP environment
        Raises:
            ForemastError, when the current env name is not a known GCP Environment
        """
        all_gcp_envs = GcpEnvironment.get_environments_from_config()
        if self.env not in all_gcp_envs:
            raise ForemastError("GCP environment %s not found in configuration", self.env)
        return all_gcp_envs[self.env]

    def cleanup(self):
        """Clean up generated files."""
        os.remove(self.raw_path)

    def check_env_defined(self):
        """Checks if the current environment is defined in the pipeline files.
        Raises a ForemastError if it is missing."""
        if not self.env:
            raise ForemastError("Environment not set")
        if self.env not in self.configs:
            raise ForemastError("Environment '{}' not found in pipeline configs.".format(self.env)
                                + "Check pipeline.json and application-master-{}.json".format(self.env))


def prepare_infrastructure():
    """Entry point for preparing the infrastructure in a specific env."""

    runner = ForemastRunner()
    runner.write_configs()

    try:
        pipeline_type = runner.configs['pipeline']['type']
        if pipeline_type is None or pipeline_type.isspace():
            raise ForemastError("pipeline.type is required")
    except KeyError:
        raise ForemastError("pipeline.type is required")

    cloud_name = configs.get_cloud_for_pipeline_type(pipeline_type)

    if cloud_name == "gcp":
        LOG.info("Will create GCP Infrastructure for pipeline.type '%s'", pipeline_type)
        prepare_infrastructure_gcp(runner)
    elif cloud_name in "aws":
        LOG.info("Will create AWS Infrastructure for pipeline.type '%s'", pipeline_type)
        prepare_infrastructure_aws(runner, pipeline_type)
    else:
        error_message = ("pipeline.type of '{0}' is not supported. "
                         "If this is a manual pipeline it is required you specify the "
                         "pipeline type in AWS_MANUAL_TYPES or GCP_MANUAL_TYPES"
                         ).format(pipeline_type)
        raise ForemastError(error_message)


def prepare_infrastructure_gcp(runner: ForemastRunner):
    """Creates GCP infrastructure for a specific env."""
    # Always create IAM, this ensure svc account and permissions is done
    runner.create_gcp_iam()


def prepare_infrastructure_aws(runner, pipeline_type):
    """Creates AWS infrastructure for a specific env."""
    runner.check_env_defined()
    archaius = runner.configs[runner.env]['app']['archaius_enabled']
    eureka = runner.configs[runner.env]['app']['eureka_enabled']

    runner.create_app()

    if pipeline_type not in ['s3', 'datapipeline', 'stepfunction']:
        runner.create_aws_iam()
        # TODO: Refactor Archaius to be fully featured
        if archaius:
            runner.create_archaius()
        runner.create_secgroups()

    if eureka:
        LOG.info("Eureka Enabled, skipping ELB and DNS setup")
    elif pipeline_type == "lambda":
        if consts.LAMBDA_STANDALONE_MODE:
            LOG.warning("Skipping Lambda creation due to LAMBDA_STANDALONE_MODE feature flag")
        else:
            LOG.info("Lambda Enabled, skipping ELB and DNS setup")
            runner.create_awslambda()
    elif pipeline_type == "s3":
        runner.create_s3app()
    elif pipeline_type == 'datapipeline':
        runner.create_datapipeline()
    elif pipeline_type == 'stepfunction':
        runner.create_aws_iam()
        runner.create_stepfunction()
    else:
        LOG.info("No Eureka, running ELB and DNS setup")
        runner.create_elb()
        runner.create_dns()

    runner.slack_notify()
    runner.cleanup()


def describe_environments(args):
    """Prints a simple visual output of environments visible to Foremast"""
    if args.parsed.cloud_provider == "gcp":
        table = get_describe_gcp_environments()
    elif args.parsed.cloud_provider == "aws":
        table = get_describe_aws_environments()
    else:
        raise ForemastError("Cannot describe cloud '{}'".format(args.parsed.cloud_provider))

    output = tabulate(table[1], table[0], tablefmt=args.parsed.print_table_format)
    if args.parsed.print_to_file:
        file = open(args.parsed.print_to_file, "w")
        file.write(output)
        file.close()
        LOG.info("Saved printed table to %s", args.parsed.print_to_file)

    print(output)


def get_describe_gcp_environments():
    """Prints a simple visual output of GCP Environments visible to Foremast
    Returns:
        tuple: first the table headers, second the table values
    """
    table_header = ["Environment", "Project", "Permitted Groups"]
    env_table = list()
    all_envs = GcpEnvironment.get_environments_from_config()
    for env_name in all_envs:
        env = all_envs[env_name]
        env_table.append([env.name, env.service_account_project, "N/A"])
        all_projects = env.get_all_projects()
        for project in all_projects:
            if "foremast_groups" in project["labels"]:
                groups = project["labels"]["foremast_groups"]
            else:
                groups = "N/A"
            env_table.append([env.name, project["projectId"], groups])
    return table_header, env_table


def get_describe_aws_environments():
    """Prints a simple visual output of AWS environments configured in Foremast
    Returns:
        tuple: first the table headers, second the table values
    """
    table_header = ["Environment"]
    env_table = list()
    for env in configs.ENVS:
        env_table.append([env])

    return table_header, env_table


def prepare_app_pipeline():
    """Entry point for application setup and initial pipeline in Spinnaker."""
    runner = ForemastRunner()
    runner.write_configs()
    runner.create_app()
    runner.create_pipeline()
    runner.cleanup()


def prepare_onetime_pipeline():
    """Entry point for single use pipeline setup in the defined app."""
    runner = ForemastRunner()
    runner.write_configs()
    runner.create_pipeline(onetime=os.getenv('ENV'))
    runner.cleanup()


def create_scaling_policy():
    """Create Auto Scaling Policy for an Auto Scaling Group."""
    runner = ForemastRunner()
    runner.write_configs()
    runner.create_autoscaling_policy()
    runner.cleanup()


def create_scheduled_actions():
    """Create Scheduled Actions for an Auto Scaling Group."""
    runner = ForemastRunner()
    runner.write_configs()
    runner.create_scheduled_actions()
    runner.cleanup()


def rebuild_pipelines(*args):
    """Entry point for rebuilding pipelines.

    Use to rebuild all pipelines or a specific group.
    """
    rebuild_all = False
    rebuild_project = os.getenv("REBUILD_PROJECT")

    if args:
        LOG.debug('Incoming arguments: %s', args)
        command_args, *_ = args
        rebuild_all = command_args.parsed.all
        rebuild_project = command_args.parsed.project

    if rebuild_project == 'ALL':
        rebuild_all = True

    if rebuild_all:
        LOG.info('Rebuilding all projects.')
    elif rebuild_project is None:
        msg = 'No REBUILD_PROJECT variable found'
        LOG.fatal(msg)
        raise SystemExit('Error: {0}'.format(msg))
    else:
        LOG.info('Rebuilding project: %s', rebuild_project)

    all_apps = utils.get_all_apps()

    for apps in all_apps:
        if 'repoProjectKey' not in apps:
            LOG.info('Skipping %s. No project key found', apps['name'])
            continue

        app_name = '{}/{}'.format(apps['repoProjectKey'], apps['repoSlug'])
        if apps['repoProjectKey'].lower() == rebuild_project.lower() or rebuild_all:
            os.environ["PROJECT"] = apps['repoProjectKey']
            os.environ["GIT_REPO"] = apps['repoSlug']
            LOG.info('Rebuilding pipelines for %s', app_name)
            runner = ForemastRunner()
            try:
                runner.write_configs()
                runner.create_pipeline()
                runner.cleanup()
            except Exception:  # pylint: disable=broad-except
                LOG.warning('Error updating pipeline for %s', app_name)


def deploy_awslambda():
    """Deploys an AWS Lambda"""
    runner = ForemastRunner()
    runner.write_configs()
    runner.deploy_awslambda()


def deploy_cloudfunction():
    """Deploys a GCP Cloud Function"""
    runner = ForemastRunner()
    runner.write_configs()
    runner.deploy_cloudfunction()


def deploy_s3app():
    """Entry point for application setup and s3 deployments"""
    runner = ForemastRunner()
    runner.write_configs()
    runner.deploy_s3app()


def promote_s3app():
    """Entry point for application setup and s3 promotions"""
    runner = ForemastRunner()
    runner.write_configs()
    runner.promote_s3app()


def debug_flag():
    """Set logging level for entry points."""
    logging.basicConfig(format=consts.LOGGING_FORMAT)

    parser = argparse.ArgumentParser(description=debug_flag.__doc__)
    add_debug(parser)
    args, _extra_args = parser.parse_known_args()

    package, *_ = __package__.split('.')
    logging.getLogger(package).setLevel(args.debug)



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Common _argparse_ arguments."""
import logging
import os

from .consts import ENVS


def add_app(parser):
    """Add an `app` flag to the _parser_."""
    parser.add_argument('-a', '--app', help='Spinnaker Application name', required=True)


def add_debug(parser):
    """Add a `debug` flag to the _parser_."""
    parser.add_argument(
        '-d', '--debug', action='store_const', const=logging.DEBUG, default=logging.INFO, help='Set DEBUG output')


def add_env(parser):
    """Add an `env` flag to the _parser_."""
    parser.add_argument(
        '-e', '--env', choices=ENVS, default=os.getenv('ENV', default='dev'), help='Deploy environment, overrides $ENV')


def add_gitlab_token(parser):
    """Add a `token-file` flag to the _parser_."""
    parser.add_argument('-t', '--token-file', help='File with GitLab API private token', default='~/.aws/git.token')


def add_properties(parser):
    """Add a `settings` flag to the _parser_."""
    parser.add_argument(
        '-p',
        '--properties',
        help='Location of JSON file produced by *create-configs*',
        default="./raw.properties.json")


def add_region(parser):
    """Add a `region` flag to the _parser_."""
    parser.add_argument('-r', '--region', default='us-east-1', help='Region to create Resources in, e.g. us-east-1')


def add_artifact_path(parser):
    """Add an `artifact-path` flag to _parser_."""
    parser.add_argument('--artifact-path', help='Local path to artifact directory for S3 deployments')


def add_artifact_version(parser):
    """Add an `artifact-version` flag to _parser_."""
    parser.add_argument('--artifact-version', help='Artifact version for S3 deployments')


def add_provider(parser):
    """Add an `provider` flag to _parser_."""
    parser.add_argument('--provider', help='Cloud provider name', default='aws')



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Destroy any S3 Resources."""
import logging

import boto3

from ...utils import get_details

LOG = logging.getLogger(__name__)


def destroy_s3(app='', env='dev', **_):
    """Destroy S3 Resources for _app_ in _env_.

    Args:
        app (str): Application name
        env (str): Deployment environment/account name

    Returns:
        boolean: True if destroyed sucessfully
    """
    session = boto3.Session(profile_name=env)
    client = session.resource('s3')

    generated = get_details(app=app, env=env)
    archaius = generated.archaius()

    bucket = client.Bucket(archaius['bucket'])

    for item in bucket.objects.filter(Prefix=archaius['path']):
        item.Object().delete()
        LOG.info('Deleted: %s/%s', item.bucket_name, item.key)

    return True



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Add application.properties to Application's S3 Bucket directory.

Help: ``python -m src.foremast.s3 -h``
"""

import argparse
import logging

from ..args import add_app, add_artifact_path, add_artifact_version, add_debug, add_env, add_properties, add_region
from ..consts import LOGGING_FORMAT
from ..utils import get_properties
from .create_archaius import init_properties
from .s3apps import S3Apps
from .s3deploy import S3Deployment

LOG = logging.getLogger(__name__)


def main():
    """Create application.properties for a given application."""
    logging.basicConfig(format=LOGGING_FORMAT)

    parser = argparse.ArgumentParser(description=main.__doc__)

    add_debug(parser)
    add_app(parser)
    add_env(parser)
    add_properties(parser)
    add_region(parser)
    add_artifact_path(parser)
    add_artifact_version(parser)

    args = parser.parse_args()

    logging.getLogger(__package__.split('.')[0]).setLevel(args.debug)

    LOG.debug('Args: %s', vars(args))

    rendered_props = get_properties(args.properties)
    if rendered_props['pipeline']['type'] == 's3':
        s3app = S3Apps(app=args.app, env=args.env, region=args.region, prop_path=args.properties)
        s3app.create_bucket()

        s3deploy = S3Deployment(
            app=args.app,
            env=args.env,
            region=args.region,
            prop_path=args.properties,
            artifact_path=args.artifact_path,
            artifact_version=args.artifact_version)
        s3deploy.upload_artifacts()
    else:
        init_properties(**vars(args))


if __name__ == '__main__':
    main()



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Test utils."""

from unittest import mock

import pytest

from foremast.exceptions import *
from foremast.utils import *


@mock.patch('foremast.utils.banners.LOG')
def test_utils_banner(mock_log):
    banner('test', border='+', width=10)
    mock_log.info.assert_called_with('+' * 10)


def test_utils_deep_chain_map():

    first = {
        'key1': {
            'subkey1': 1,
        },
    }
    second = {
        'key1': {
            'subkey2': 2,
        },
    }

    result = {
        'key1': {
            'subkey1': 1,
            'subkey2': 2,
        }
    }

    assert DeepChainMap(first, second) == result
    with pytest.raises(KeyError):
        assert DeepChainMap(first, second)['key2'] == result


def test_utils_pipeline_check_managed():

    assert check_managed_pipeline('app [onetime]', 'app') == 'onetime'
    assert check_managed_pipeline('app [us-east-1]', 'app') == 'us-east-1'

    bad_names = [
        'something',
        'app',
        'app [us-east-1',
        'app us-east-1]',
        'app [us-east-1',
        'app us-east-1]',
        'app name',
        'app2 [us-east-1]',
        'app name [us-east-1]',
    ]

    for name in bad_names:
        with pytest.raises(ValueError):
            check_managed_pipeline(name=name, app_name='app')


@mock.patch('foremast.utils.pipelines.gate_request')
def test_utils_pipeline_get_all_pipelines(mock_gate_request):
    mock_gate_request.return_value.json.return_value = {}
    result = get_all_pipelines(app='app')
    assert result == {}


@mock.patch('foremast.utils.pipelines.get_all_pipelines')
def test_utils_pipeline_get_pipeline_id(mock_get_pipelines):
    """Verify Pipeline ID response."""
    data = [
        {
            'name': 'app',
            'id': 100
        },
    ]
    mock_get_pipelines.return_value = data

    result = get_pipeline_id(app='test', name='app')
    mock_get_pipelines.assert_called_once_with(app='test')
    assert result is 100

    result = get_pipeline_id(app='embarrassingly', name='badapp')
    mock_get_pipelines.assert_called_with(app='embarrassingly')
    assert result == None


def test_utils_generate_packer_filename():
    a = generate_packer_filename('aws', 'us-east-1', 'chroot')
    assert a == 'aws_us-east-1_chroot.json'


@mock.patch('foremast.utils.elb.gate_request')
def test_utils_find_elb(gate_request_mock):
    results = [{'account': 'dev', 'region': 'us-east-1', 'dnsname': 'appdns'}]
    gate_request_mock.return_value.json.return_value = results
    a = find_elb('app', 'dev', 'us-east-1')
    assert a == 'appdns'

    with pytest.raises(SpinnakerElbNotFound):
        # we already filter by app, so sending incorrect env/region combo
        # will trigger the error
        find_elb('app', 'devbad', 'us-east-1')


@mock.patch('foremast.utils.slack.slacker')
def test_utils_post_slack_message(mock_slack):
    post_slack_message('test', '#test')
    mock_slack.called


@mock.patch('foremast.utils.apps.gate_request')
def test_utils_apps_get_details(mock_gate_request):
    data = {'attributes': {'repoProjectKey': 'group', 'repoSlug': 'repo1'}}
    mock_gate_request.return_value.json.return_value = data

    result = get_details(app='repo1group', env='dev')
    assert result.app_name() == 'repo1group'

    with pytest.raises(SpinnakerAppNotFound):
        mock_gate_request.return_value.ok = False
        result = get_details(app='repo1group', env='dev')
        assert result.app_name() == 'repo1group'


@mock.patch('foremast.utils.apps.gate_request')
def test_utils_apps_get_all_apps(mock_gate_request):
    data = []
    mock_gate_request.return_value.json.return_value = data

    result = get_all_apps()
    assert result == []

    with pytest.raises(AssertionError):
        mock_gate_request.return_value.ok = False
        result = get_all_apps()


@mock.patch('foremast.utils.dns.boto3.Session')
@mock.patch('foremast.utils.dns.DOMAIN', 'test')
def test_utils_dns_get_zone_ids(mock_boto3):
    data = {
        'HostedZones': [
            {
                'Name': 'internal.example.com',
                'Id': 100,
                'Config': {
                    'PrivateZone': True
                }
            },
            {
                'Name': 'external.example.com',
                'Id': 101,
                'Config': {
                    'PrivateZone': False
                }
            },
        ]
    }

    data_external = {
        'HostedZones': [
            {
                'Name': 'internal.example.com',
                'Id': 100,
                'Config': {
                    'PrivateZone': False
                }
            },
            {
                'Name': 'external.example.com',
                'Id': 101,
                'Config': {
                    'PrivateZone': False
                }
            },
        ]
    }

    mock_boto3.return_value.client.return_value.list_hosted_zones_by_name.return_value = data

    # default case
    result = get_dns_zone_ids()
    assert result == [100]

    # all zones
    result = get_dns_zone_ids(facing='external')
    assert result == [100, 101]

    # all internal
    result = get_dns_zone_ids(facing='internal')
    assert result == [100]

    # unkown param - mixed zones
    result = get_dns_zone_ids(facing='wrong_param')
    assert result == [100]

    # no internal zones
    mock_boto3.return_value.client.return_value.list_hosted_zones_by_name.return_value = data_external
    result = get_dns_zone_ids(facing='internal')
    assert result == []

    # unknown param - no internal zones
    result = get_dns_zone_ids(facing='wrong_param')
    assert result == []


@mock.patch('foremast.utils.dns.boto3.Session')
def test_find_existing_record(mock_session):
    """Check that a record is found correctly"""

    dns_values = {'env': 'dev', 'zone_id': '/hostedzone/TESTTESTS279', 'dns_name': 'test.example.com'}
    test_records = [{
        'ResourceRecordSets': [{
            'Name': 'test.example.com.',
            'Type': 'CNAME'
        }]
    }, {
        'ResourceRecordSets': [{
            'Name': 'test.example.com.',
            'Failover': 'PRIMARY'
        }]
    }, {
        'ResourceRecordSets': [{
            'Name': 'test.example.com.',
            'Type': 'A'
        }]
    }]
    client = mock_session.return_value.client.return_value
    client.get_paginator.return_value.paginate.return_value = test_records
    assert find_existing_record(
        dns_values['env'], dns_values['zone_id'], dns_values['dns_name'], check_key='Type', check_value='CNAME') == {
            'Name': 'test.example.com.',
            'Type': 'CNAME'
        }
    assert find_existing_record(
        dns_values['env'], dns_values['zone_id'], dns_values['dns_name'], check_key='Failover',
        check_value='PRIMARY') == {
            'Name': 'test.example.com.',
            'Failover': 'PRIMARY'
        }
    assert find_existing_record(
        dns_values['env'], dns_values['zone_id'], 'bad.example.com', check_key='Type', check_value='CNAME') == None


@mock.patch('foremast.utils.security_group.gate_request')
@mock.patch('foremast.utils.security_group.get_vpc_id')
def test_utils_sg_get_security_group_id(mock_vpc_id, mock_gate_request):
    data = {'id': 100}
    mock_gate_request.return_value.json.return_value = data

    # default - happy path
    result = get_security_group_id()
    assert result == 100

    # security group not found
    with pytest.raises(SpinnakerSecurityGroupError):
        mock_gate_request.return_value.json.return_value = {}
        result = get_security_group_id()

    # error getting details
    with pytest.raises(AssertionError):
        mock_gate_request.return_value.ok = False
        result = get_security_group_id()


@mock.patch('foremast.utils.vpc.gate_request')
def test_utils_vpc_get_vpc_id(mock_gate_request):
    data = [
        {
            'id': 100,
            'name': 'vpc',
            'account': 'dev',
            'region': 'us-east-1'
        },
    ]
    mock_gate_request.return_value.json.return_value = data

    # default - happy path
    result = get_vpc_id(account='dev', region='us-east-1')
    assert result == 100

    # vpc not found
    with pytest.raises(SpinnakerVPCIDNotFound):
        result = get_vpc_id(account='dev', region='us-west-2')
        assert result == 100

    # error getting details
    with pytest.raises(SpinnakerVPCNotFound):
        mock_gate_request.return_value.ok = False
        result = get_vpc_id(account='dev', region='us-east-1')


SUBNET_DATA = [
    {
        'vpcId': 100,
        'account': 'dev',
        'id': 1,
        'purpose': 'internal',
        'region': 'us-east-1',
        'target': 'ec2',
        'availabilityZone': []
    },
    {
        'vpcId': 101,
        'account': 'dev',
        'id': 2,
        'purpose': 'other',
        'region': 'us-west-2',
        'target': 'ec2',
        'availabilityZone': ['us-west-2a', 'us-west-2b']
    },
]


@mock.patch('foremast.utils.subnets.gate_request')
def test_utils_subnets_get_subnets(mock_gate_request):
    """Find one subnet."""
    mock_gate_request.return_value.json.return_value = SUBNET_DATA

    # default - happy path
    result = get_subnets(env='dev', region='us-east-1')
    assert result == {
        'subnet_ids': {
            'us-east-1': [SUBNET_DATA[0]['id']],
        },
        'us-east-1': [[]],
    }


@mock.patch('foremast.utils.subnets.gate_request')
def test_utils_subnets_get_subnets_multiple_az(mock_gate_request):
    """Find multiple Availability Zones."""
    mock_gate_request.return_value.json.return_value = SUBNET_DATA

    # default - happy path w/multiple az
    result = get_subnets(env='dev', region='')
    assert result == {'dev': {'us-west-2': [['us-west-2a', 'us-west-2b']], 'us-east-1': [[]]}}


@mock.patch('foremast.utils.subnets.gate_request')
def test_utils_subnets_get_subnets_subnet_not_found(mock_gate_request):
    """Trigger SpinnakerSubnetError when no subnets found."""
    mock_gate_request.return_value.json.return_value = SUBNET_DATA

    # subnet not found
    with pytest.raises(SpinnakerSubnetError):
        result = get_subnets(env='dev', region='us-west-1')
        assert result == {'us-west-1': [[]]}


@mock.patch('foremast.utils.subnets.gate_request')
def test_utils_subnets_get_subnets_api_error(mock_gate_request):
    """Trigger SpinnakerTimeout when API has error."""
    mock_gate_request.return_value.json.return_value = SUBNET_DATA

    # error getting details
    with pytest.raises(SpinnakerTimeout):
        mock_gate_request.return_value.ok = False
        result = get_subnets()


@mock.patch('foremast.utils.tasks.check_task')
@mock.patch('foremast.utils.tasks.post_task')
@mock.patch('foremast.utils.tasks.TASK_TIMEOUTS')
def test_utils_timeout_per_env(mock_check_task, mock_requests_post, mock_timeouts):
    """Verify custom timeout propagates to check_task"""
    mock_requests_post.return_value = 5
    task_data = {"job": [{"credentials": "dev", "type": "fake_task"}]}
    mock_timeouts.side_effect = {"dev": {"fake_task": "240"}}
    tasks.wait_for_task(task_data)
    assert mock_check_task.called_with("fake_task", 240)
    assert mock_check_task.called_with("fake_task", tasks.DEFAULT_TASK_TIMEOUT)


@mock.patch('foremast.utils.tasks.check_task')
@mock.patch('foremast.utils.tasks.post_task')
@mock.patch('foremast.utils.tasks.TASK_TIMEOUTS')
def test_utils_default_timeout(mock_check_task, mock_requests_post, mock_timeouts):
    """default timeout for tasks is applied if missing from timeout data"""
    mock_requests_post.return_value = 5
    task_data = {"job": [{"credentials": "dev", "type": "really_fake_task"}]}
    mock_timeouts.side_effect = {"dev": {"fake_task": "240"}}
    tasks.wait_for_task(task_data)
    assert mock_check_task.called_with("really_fake_task", tasks.DEFAULT_TASK_TIMEOUT)



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Create S3 lambda event"""

import json
import logging

import boto3

from ...utils import add_lambda_permissions, get_lambda_alias_arn, get_template

LOG = logging.getLogger(__name__)


def create_s3_event(app_name, env, region, bucket, triggers):
    """Create S3 lambda events from triggers

    Args:
        app_name (str): name of the lambda function
        env (str): Environment/Account for lambda function
        region (str): AWS region of the lambda function
        triggers (list): List of triggers from the settings
    """
    session = boto3.Session(profile_name=env, region_name=region)
    s3_client = session.client('s3')

    lambda_alias_arn = get_lambda_alias_arn(app_name, env, region)

    LOG.debug("Lambda ARN for lambda function %s is %s.", app_name, lambda_alias_arn)
    LOG.debug("Creating S3 events for bucket %s", bucket)

    # allow lambda trigger permission from bucket
    principal = 's3.amazonaws.com'
    statement_id = "{}_s3_{}".format(app_name, bucket).replace('.', '')
    source_arn = "arn:aws:s3:::{}".format(bucket)
    add_lambda_permissions(
        function=lambda_alias_arn,
        env=env,
        region=region,
        principal=principal,
        statement_id=statement_id,
        source_arn=source_arn)

    # configure events on s3 bucket to trigger lambda function
    template_kwargs = {"lambda_arn": lambda_alias_arn, "triggers": triggers}
    config = get_template(template_file='infrastructure/lambda/s3_event.json.j2', **template_kwargs)

    # Get Bucket Tags to check if buckets are locked down from automated attachment
    bucket_tag_response = s3_client.get_bucket_tagging(Bucket=bucket)
    bucket_tags = bucket_tag_response['TagSet']
    LOG.debug('Bucket Tags: %s', bucket_tags)

    disable_lambda_trigger_tag = {'Key': 'FOREMAST_LAMBDA_RESTRICT_BUCKET_NOTIFICATIONS', 'Value': 'true'}

    if disable_lambda_trigger_tag in bucket_tags:
        LOG.warning('WARN: Target S3 bucket limits bucket events. Check bucket for tag %s', disable_lambda_trigger_tag)
    else:
        s3_client.put_bucket_notification_configuration(Bucket=bucket, NotificationConfiguration=json.loads(config))
        LOG.info("Created lambda %s S3 event on bucket %s", app_name, bucket)



#   Foremast - Pipeline Tooling
#
#   Copyright 2020 Redbox Automated Retail, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Set of utility functions for Foremasts configs"""


def verify_approval_skip(data, env, env_configs):
    """Determines if a approval stage can be added/removed from a given
    environment pipeline stage based on environment setting. Defaults to false,
    and verifies administrators allow skips in given environments.
    Args:
        data (dict): environment config data from pipeline files
        env (str): Name of environment
        env_configs (dict): environment configs from foremast files

    Returns:
        bool: result of approval skip check/verification and setting
    """

    approval_skip = False

    if 'approval_skip' in data['app']:
        if env in env_configs and env_configs[env]['enable_approval_skip']:
            approval_skip = data['app']['approval_skip']

    return approval_skip



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Create Security Groups for Spinnaker Pipelines.

Security Group port specifications will be sourced from the `application.json`
files for each environment.

Examples:
    application-master.json::

        {
            "security_group": {
                "description": "Security Group description",
                "ingress": {
                    "eureka": [
                        {"start_port": 80, "end_port": 8080, "protocol": "tcp"}
                    ],
                    "coreforrest": [
                        8080,
                        8443
                    ],
                    "0.0.0.0/0": [
                        8080
                    ]
                }
            }
        }
"""
import ipaddress
import logging
from contextlib import suppress

import boto3
from boto3.exceptions import botocore
from deepmerge import conservative_merger

from ..consts import DEFAULT_SECURITYGROUP_RULES
from ..exceptions import (ForemastConfigurationFileError, SpinnakerSecurityGroupCreationFailed,
                          SpinnakerSecurityGroupError)
from ..utils import get_details, get_properties, get_security_group_id, get_template, get_vpc_id, wait_for_task


class SpinnakerSecurityGroup:
    """Manipulate Spinnaker Security Groups.

    Args:
        app (str): Application name.
        env (str): Deployment environment.
        prop_path (str): Path to the raw.properties.json.
        region (str): AWS Region.
    """

    def __init__(self, app=None, env=None, region=None, prop_path=None):
        self.log = logging.getLogger(__name__)

        self.app_name = app
        self.env = env
        self.region = region

        self.properties = get_properties(properties_file=prop_path, env=self.env, region=self.region)
        self.generated = get_details(app=self.app_name)
        self.group = self.generated.data['project']

    def _validate_cidr(self, rule):
        """Validate the cidr block in a rule.

        Returns:
            True: Upon successful completion.

        Raises:
            SpinnakerSecurityGroupCreationFailed: CIDR definition is invalid or
                the network range is too wide.
        """
        try:
            network = ipaddress.IPv4Network(rule['app'])
        except (ipaddress.NetmaskValueError, ValueError) as error:
            raise SpinnakerSecurityGroupCreationFailed(error)

        self.log.debug('Validating CIDR: %s', network.exploded)

        return True

    def _process_rules(self, rules):
        """Process rules into cidr and non-cidr lists.

        Args:
            rules (list): Allowed Security Group ports and protocols.

        Returns:
            (list, list): Security Group reference rules and custom CIDR rules.
        """
        cidr = []
        non_cidr = []

        for rule in rules:
            if '.' in rule['app']:
                self.log.debug('Custom CIDR rule: %s', rule)
                self._validate_cidr(rule)
                cidr.append(rule)
            else:
                self.log.debug('SG reference rule: %s', rule)
                non_cidr.append(rule)

        self.log.debug('Custom CIDR rules: %s', cidr)
        self.log.debug('SG reference rules: %s', non_cidr)
        return non_cidr, cidr

    def add_tags(self):
        """Add tags to security group.

        Returns:
            True: Upon successful completion.
        """
        session = boto3.session.Session(profile_name=self.env, region_name=self.region)
        resource = session.resource('ec2')
        group_id = get_security_group_id(self.app_name, self.env, self.region)
        security_group = resource.SecurityGroup(group_id)

        try:
            tag = security_group.create_tags(
                DryRun=False,
                Tags=[{
                    'Key': 'app_group',
                    'Value': self.group
                }, {
                    'Key': 'app_name',
                    'Value': self.app_name
                }])
            self.log.debug('Security group has been tagged: %s', tag)
        except botocore.exceptions.ClientError as error:
            self.log.warning(error)

        return True

    def add_cidr_rules(self, rules):
        """Add cidr rules to security group via boto.

        Args:
            rules (list): Allowed Security Group ports and protocols.

        Returns:
            True: Upon successful completion.

        Raises:
            SpinnakerSecurityGroupError: boto3 call failed to add CIDR block to
                Security Group.
        """
        session = boto3.session.Session(profile_name=self.env, region_name=self.region)
        client = session.client('ec2')

        group_id = get_security_group_id(self.app_name, self.env, self.region)

        for rule in rules:
            data = {
                'DryRun':
                False,
                'GroupId':
                group_id,
                'IpPermissions': [{
                    'IpProtocol': rule['protocol'],
                    'FromPort': rule['start_port'],
                    'ToPort': rule['end_port'],
                    'IpRanges': [{
                        'CidrIp': rule['app']
                    }]
                }]
            }
            self.log.debug('Security Group rule: %s', data)

            try:
                client.authorize_security_group_ingress(**data)
            except botocore.exceptions.ClientError as error:
                if 'InvalidPermission.Duplicate' in str(error):
                    self.log.debug('Duplicate rule exist, that is OK.')
                else:
                    msg = 'Unable to add cidr rules to {}'.format(rule.get('app'))
                    self.log.error(msg)
                    raise SpinnakerSecurityGroupError(msg)

        return True

    def resolve_self_references(self, rules):
        """Resolves `$self` references to actual application name in security group rules."""
        with suppress(KeyError):
            rule = rules.pop('$self')
            rules[self.app_name] = rule
        return rules

    def update_default_rules(self):
        """Concatinate application and global security group rules."""
        app_ingress = self.properties['security_group']['ingress']
        ingress = conservative_merger.merge(DEFAULT_SECURITYGROUP_RULES, app_ingress)
        resolved_ingress = self.resolve_self_references(ingress)
        self.log.info('Updated default rules:\n%s', ingress)
        return resolved_ingress

    def _create_security_group(self, ingress):
        """Send a POST to spinnaker to create a new security group.

        Returns:
            boolean: True if created successfully

        """
        template_kwargs = {
            'app': self.app_name,
            'env': self.env,
            'region': self.region,
            'vpc': get_vpc_id(self.env, self.region),
            'description': self.properties['security_group']['description'],
            'ingress': ingress,
        }

        secgroup_json = get_template(
            template_file='infrastructure/securitygroup_data.json.j2', formats=self.generated, **template_kwargs)

        wait_for_task(secgroup_json)
        return True

    def create_security_group(self):  # noqa
        """Send a POST to spinnaker to create or update a security group.

        Returns:
            boolean: True if created successfully

        Raises:
            ForemastConfigurationFileError: Missing environment configuration or
                misconfigured Security Group definition.
        """
        ingress_rules = []

        try:
            security_id = get_security_group_id(name=self.app_name, env=self.env, region=self.region)
        except (SpinnakerSecurityGroupError, AssertionError):
            self._create_security_group(ingress_rules)
        else:
            self.log.debug('Security Group ID %s found for %s.', security_id, self.app_name)

        try:
            ingress = self.update_default_rules()
        except KeyError:
            msg = 'Possible missing configuration for "{0}".'.format(self.env)
            self.log.error(msg)
            raise ForemastConfigurationFileError(msg)

        for app in ingress:
            rules = ingress[app]

            # Essentially we have two formats: simple, advanced
            # - simple: is just a list of ports
            # - advanced: selects ports ranges and protocols
            for rule in rules:
                ingress_rule = self.create_ingress_rule(app, rule)
                ingress_rules.append(ingress_rule)

        ingress_rules_no_cidr, ingress_rules_cidr = self._process_rules(ingress_rules)

        self._create_security_group(ingress_rules_no_cidr)

        # Append cidr rules
        self.add_cidr_rules(ingress_rules_cidr)

        # Tag security group
        self.add_tags()

        self.log.info('Successfully created %s security group', self.app_name)
        return True

    def create_ingress_rule(self, app, rule):
        """Create a normalized ingress rule.

        Args:
            app (str): Application name
            rule (dict or int): Allowed Security Group ports and protocols.

        Returns:
            dict: Contains app, start_port, end_port, protocol, cross_account_env and cross_account_vpc_id

        """
        if isinstance(rule, dict):
            # Advanced
            start_port = rule.get('start_port')
            end_port = rule.get('end_port')
            protocol = rule.get('protocol', 'tcp')

            requested_cross_account = rule.get('env', self.env)
            if self.env == requested_cross_account:
                # We are trying to use cross-account security group settings within the same account
                # We should not allow this.
                cross_account_env = None
                cross_account_vpc_id = None
            else:
                cross_account_env = requested_cross_account
                cross_account_vpc_id = get_vpc_id(cross_account_env, self.region)

        else:
            start_port = rule
            end_port = rule
            protocol = 'tcp'
            cross_account_env = None
            cross_account_vpc_id = None

        created_rule = {
            'app': app,
            'start_port': start_port,
            'end_port': end_port,
            'protocol': protocol,
            'cross_account_env': cross_account_env,
            'cross_account_vpc_id': cross_account_vpc_id
        }
        self.log.debug('Normalized ingress rule: %s', created_rule)
        return created_rule



"""Foremast CLI commands."""
import argparse
import collections
import logging
import os

from . import runner, validate
from .args import add_debug, add_env
from .consts import LOGGING_FORMAT, SHORT_LOGGING_FORMAT
from .version import print_version

LOG = logging.getLogger(__name__)


def add_infra(subparsers):
    """Infrastructure subcommands."""
    infra_parser = subparsers.add_parser('infra', help=runner.prepare_infrastructure.__doc__)
    infra_parser.set_defaults(func=runner.prepare_infrastructure)


def add_pipeline(subparsers):
    """Pipeline subcommands."""
    pipeline_parser = subparsers.add_parser(
        'pipeline', help=add_pipeline.__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    pipeline_parser.set_defaults(func=pipeline_parser.print_help)

    pipeline_subparsers = pipeline_parser.add_subparsers(title='Pipelines')

    pipeline_full_parser = pipeline_subparsers.add_parser(
        'app', help=runner.prepare_app_pipeline.__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    pipeline_full_parser.set_defaults(func=runner.prepare_app_pipeline)

    pipeline_onetime_parser = pipeline_subparsers.add_parser(
        'onetime', help=runner.prepare_onetime_pipeline.__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    pipeline_onetime_parser.set_defaults(func=runner.prepare_onetime_pipeline)
    add_env(pipeline_onetime_parser)


def add_cloudfunctions(subparsers):
    """CloudFunctions subcommands."""
    # Add subparser like `foremast cloudfunctions`
    cloudfunctions_parser = subparsers.add_parser('cloudfunctions', help="Options for deploying GCP Cloud Functions")

    # Add more subparsers, like foremast cloudfunctions deploy`
    cf_subparsers = cloudfunctions_parser.add_subparsers(title='GCP Cloud Functions Subcommands')
    deploy_parser = cf_subparsers.add_parser("deploy", help=runner.deploy_cloudfunction.__doc__)
    deploy_parser.set_defaults(func=runner.deploy_cloudfunction)


def add_lambdas(subparsers):
    """AWS Lambdas subcommands."""
    # Add subparser like `foremast lambdas`
    lambdas_parser = subparsers.add_parser('lambdas', help="Options for deploying AWS Lambdas")

    # Add more subparsers, like foremast lambdas deploy`
    lambda_subparsers = lambdas_parser.add_subparsers(title='AWS Lambdas Subcommands')
    deploy_parser = lambda_subparsers.add_parser("deploy", help=runner.deploy_awslambda.__doc__)
    deploy_parser.set_defaults(func=runner.deploy_awslambda)


def add_rebuild(subparsers):
    """Rebuild Pipeline subcommands."""
    rebuild_parser = subparsers.add_parser(
        'rebuild', help=runner.rebuild_pipelines.__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    rebuild_parser.set_defaults(func=runner.rebuild_pipelines)
    rebuild_parser.add_argument('-a', '--all', action='store_true', help='Rebuild all Pipelines')
    rebuild_parser.add_argument(
        'project',
        nargs='?',
        default=os.getenv('REBUILD_PROJECT'),
        help='Project to rebuild, overrides $REBUILD_PROJECT')


def add_autoscaling(subparsers):
    """Auto Scaling Group Policy subcommands."""
    autoscaling_parser = subparsers.add_parser(
        'autoscaling',
        help=runner.create_scaling_policy.__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    autoscaling_parser.set_defaults(func=runner.create_scaling_policy)


def add_scheduled_actions(subparsers):
    """Auto Scaling Group Scheduled Actions subcommands."""
    scheduled_actions_parser = subparsers.add_parser(
        'scheduledactions',
        help=runner.create_scheduled_actions.__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    scheduled_actions_parser.set_defaults(func=runner.create_scheduled_actions)


def add_validate(subparsers):
    """Validate Spinnaker setup."""
    validate_parser = subparsers.add_parser(
        'validate', help=add_validate.__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    validate_parser.set_defaults(func=validate_parser.print_help)

    validate_subparsers = validate_parser.add_subparsers(title='Testers')

    validate_all_parser = validate_subparsers.add_parser(
        'all', help=validate.validate_all.__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    validate_all_parser.set_defaults(func=validate.validate_all)

    validate_gate_parser = validate_subparsers.add_parser(
        'gate', help=validate.validate_gate.__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    validate_gate_parser.set_defaults(func=validate.validate_gate)


def add_describe(subparsers):
    """Describe subcommands"""
    describe_parser = subparsers.add_parser('describe', help="Shows details of specific Foremast "
                                                             "resources/configurations")
    desc_parser_subparsers = describe_parser.add_subparsers(title='Describe Commands', description='Available commands')
    add_describe_environments(desc_parser_subparsers)


def add_describe_environments(describe_subparser):
    desc_env_parser = describe_subparser.add_parser("environments", help="Shows IAM details for a cloud provider")
    desc_env_parser.set_defaults(func=runner.describe_environments)
    desc_env_parser.add_argument('cloud_provider', choices=['aws', 'gcp'], help="Cloud provider to describe")
    desc_env_parser.add_argument('--print-table-format', default="fancy_grid", required=False,
                                 help="Format of output.  Common options are grid, fancy_grid, jira."
                                 "See tabulate package docs for full list of options.")
    desc_env_parser.add_argument('--print-to-file', default=None, required=False,
                                 help="File path to output the printed table to instead of displaying on the console")


def main(manual_args=None):
    """Foremast, your ship's support. """
    parser = argparse.ArgumentParser(description=main.__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.set_defaults(func=parser.print_help)
    add_debug(parser)
    parser.add_argument(
        '-s',
        '--short-log',
        action='store_const',
        const=SHORT_LOGGING_FORMAT,
        default=LOGGING_FORMAT,
        help='Truncated logging format')
    parser.add_argument('-v', '--version', action='store_true', help=print_version.__doc__)

    subparsers = parser.add_subparsers(title='Commands', description='Available activies')

    add_infra(subparsers)
    add_pipeline(subparsers)
    add_rebuild(subparsers)
    add_cloudfunctions(subparsers)
    add_lambdas(subparsers)
    add_autoscaling(subparsers)
    add_scheduled_actions(subparsers)
    add_validate(subparsers)
    add_describe(subparsers)

    CliArgs = collections.namedtuple('CliArgs', ['parsed', 'extra'])

    parsed, extra = parser.parse_known_args(args=manual_args)
    args = CliArgs(parsed, extra)

    logging.basicConfig(format=args.parsed.short_log)

    package, *_ = __package__.split('.')
    logging.getLogger(package).setLevel(args.parsed.debug)

    LOG.debug('Arguments: %s', args)

    if args.parsed.version:
        args.parsed.func = print_version

    try:
        args.parsed.func(args)
    except (AttributeError, TypeError):
        args.parsed.func()


if __name__ == '__main__':
    main()



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Notify Slack channel."""
import time

from ..utils import get_properties, get_template, post_slack_message


class SlackNotification:
    """Post slack notification.
    Inform users about infrastructure changes to prod* accounts.

    Args:
        app (str): Application name
        env (str): Environment/account name of changed infrastructure
        prop_path (str): Path to the rendered configuration files
    """

    def __init__(self, app=None, env=None, prop_path=None):
        timestamp = time.strftime("%B %d, %Y %H:%M:%S %Z", time.gmtime())

        self.settings = get_properties(prop_path)
        short_commit_sha = self.settings['pipeline']['config_commit'][0:11]

        self.info = {
            'app': app,
            'env': env,
            'config_commit_short': short_commit_sha,
            'timestamp': timestamp,
        }

    def post_message(self):
        """Send templated message to **#deployments-{env}**.

        Primarily for production deployments.
        """
        message = get_template(template_file='slack/pipeline-prepare-ran.j2', info=self.info)
        channel = '#deployments-prod'
        post_slack_message(message=message, channel=channel, username='pipeline-bot', icon_emoji=':gear:')

    def notify_slack_channel(self):
        """Post message to a defined Slack channel."""
        message = get_template(template_file='slack/pipeline-prepare-ran.j2', info=self.info)

        if self.settings['pipeline']['notifications']['slack']:
            post_slack_message(
                message=message,
                channel=self.settings['pipeline']['notifications']['slack'],
                username='pipeline-bot',
                icon_emoji=':gear:')



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Generate base64 encoded User Data."""
import base64

from .templates import get_template


def generate_encoded_user_data(
        env='dev',
        region='us-east-1',
        generated=None,
        group_name='',
        pipeline_type='',
        canary=False,
):
    r"""Generate base64 encoded User Data.

    Args:
        env (str): Deployment environment, e.g. dev, stage.
        region (str): AWS Region, e.g. us-east-1.
        generated (foremastutils.Generator): Generated naming formats.
        group_name (str): Application group nane, e.g. core.
        pipeline_type (str): Type of Foremast Pipeline to configure.

    Returns:
        str: base64 encoded User Data script.

            #!/bin/bash
            export CLOUD_ENVIRONMENT=dev
            export CLOUD_ENVIRONMENT_C=dev
            export CLOUD_ENVIRONMENT_P=dev
            export CLOUD_ENVIRONMENT_S=dev
            export CLOUD_APP=coreforrest
            export CLOUD_APP_GROUP=forrest
            export CLOUD_STACK=forrest
            export EC2_REGION=us-east-1
            export CLOUD_DOMAIN=dev.example.com
            printenv | grep 'CLOUD\|EC2' | awk '$0="export "$0'>> /etc/foremast/cloud_env

    """
    # We need to handle the case of prodp and prods for different URL generation
    if env in ["prod", "prodp", "prods"]:
        env_c, env_p, env_s = "prod", "prodp", "prods"
    else:
        env_c, env_p, env_s = env, env, env

    user_data = get_template(
        template_file='infrastructure/user_data.sh.j2',
        env=env,
        env_c=env_c,
        env_p=env_p,
        env_s=env_s,
        region=region,
        app_name=generated.app_name(),
        group_name=group_name,
        pipeline_type=pipeline_type,
        canary=canary,
        formats=generated,
    )
    return base64.b64encode(user_data.encode()).decode()



#   Foremast - Pipeline Tooling
#
#   Copyright 2019 Redbox Automated Retail, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Create Event Source Mapping trigger for lambda"""

import logging

import boto3

from datetime import datetime
from ...exceptions import DynamoDBTableNotFound
from ...utils import get_dynamodb_stream_arn, get_lambda_alias_arn

LOG = logging.getLogger(__name__)


def create_event_source_mapping_trigger(app_name, env, region, event_source, rules): # noqa: MC0001
    """Create event source mapping trigger from rules for AWS Lambda.

    Args:
        app_name (str): name of the lambda function
        env (str): Environment/Account for lambda function
        region (str): AWS region of the lambda function
        event_source (str): Event Source Service (DynamoDB, Kinesis, SQS)
        rules (str): Trigger rules from the settings
    """
    session = boto3.Session(profile_name=env, region_name=region)
    lambda_client = session.client('lambda')
    event_defaults = {
        'dynamodb-stream': {
            'service_name': 'DynamoDB Stream',
            'batch_size': 100,
            'batch_window': 0,
            'starting_position': 'TRIM_HORIZON',
            'split_on_error': False,
            'max_retry_attempts': 10000,
            'destination_config': {},
            'max_record_age': 604800
        },
        'kinesis-stream': {
            'service_name': 'Kinesis Stream',
            'batch_size': 100,
            'batch_window': 0,
            'parallelization_factor': 1,
            'starting_position': 'TRIM_HORIZON',
            'starting_position_timestamp': 0,
            'split_on_error': False,
            'max_retry': 10000,
            'destination_config': {},
            'max_record_age': 604800
        },
        'sqs': {
            'service_name': 'SQS Queue',
            'batch_size': 10
        }
    }

    if event_source == 'dynamodb-stream':
        trigger_arn = rules.get('stream_arn')
        if not trigger_arn:
            trigger_arn = rules.get('table_arn')
            if not trigger_arn:
                raise DynamoDBTableNotFound

        event_source_arn = get_dynamodb_stream_arn(arn_string=trigger_arn, account=env, region=region)
    elif event_source == 'kinesis-stream':
        event_source_arn = rules.get('stream_arn')
    elif event_source == 'sqs':
        event_source_arn = rules.get('queue_arn')
    else:
        raise NotImplementedError

    lambda_alias_arn = get_lambda_alias_arn(app=app_name, account=env, region=region)

    event_sources = lambda_client.list_event_source_mappings(FunctionName=lambda_alias_arn)
    LOG.debug('Found event sources: {0}'.format(event_sources))

    for each_source in event_sources['EventSourceMappings']:
        if each_source['EventSourceArn'] == event_source_arn and event_source == 'dynamodb-stream':
            event_uuid = each_source['UUID']
            lambda_client.update_event_source_mapping(
                UUID=event_uuid,
                FunctionName=lambda_alias_arn,
                BatchSize=rules.
                get('batch_size', event_defaults[event_source]['batch_size']),
                MaximumBatchingWindowInSeconds=rules.
                get('batch_window', event_defaults[event_source]['batch_window']),
                BisectBatchOnFunctionError=rules.
                get('split_on_error', event_defaults[event_source]['split_on_error']),
                MaximumRetryAttempts=rules.
                get('max_retry', event_defaults[event_source]['max_retry']),
                DestinationConfig=rules.
                get('destination_config', event_defaults[event_source]['destination_config']),
                MaximumRecordAgeInSeconds=rules.
                get('max_record_age', event_defaults[event_source]['max_record_age']))
            LOG.debug('{0} event trigger updated'.format(event_defaults[event_source]['service_name']))
            break
        elif each_source['EventSourceArn'] == event_source_arn and event_source == 'kinesis-stream':
            event_uuid = each_source['UUID']
            lambda_client.update_event_source_mapping(
                UUID=event_uuid,
                FunctionName=lambda_alias_arn,
                BatchSize=rules.
                get('batch_size', event_defaults[event_source]['batch_size']),
                BisectBatchOnFunctionError=rules.
                get('split_on_error', event_defaults[event_source]['split_on_error']),
                DestinationConfig=rules.
                get('destination_config', event_defaults[event_source]['destination_config']),
                MaximumBatchingWindowInSeconds=rules.
                get('batch_window', event_defaults[event_source]['batch_window']),
                MaximumRetryAttempts=rules.
                get('max_retry', event_defaults[event_source]['max_retry']),
                MaximumRecordAgeInSeconds=rules.
                get('max_record_age', event_defaults[event_source]['max_record_age']),
                ParallelizationFactor=rules.
                get('parallelization_factor', event_defaults[event_source]['parallelization_factor']))
            LOG.debug('{0} event trigger updated'.format(event_defaults[event_source]['service_name']))
            break
        else:
            event_uuid = each_source['UUID']
            lambda_client.update_event_source_mapping(
                UUID=event_uuid,
                FunctionName=lambda_alias_arn,
                BatchSize=rules.get('batch_size', event_defaults[event_source]['batch_size']))
            LOG.debug('{0} event trigger updated'.format(event_defaults[event_source]['service_name']))
            break
    else:
        if event_source == 'sqs':
            lambda_client.create_event_source_mapping(
                EventSourceArn=event_source_arn,
                FunctionName=lambda_alias_arn,
                BatchSize=rules.get('batch_size', event_defaults[event_source]['batch_size']))
        else:
            starting_position = rules.get('starting_position', event_defaults[event_source]['starting_position'])
            if starting_position == 'AT_TIMESTAMP':
                lambda_client.create_event_source_mapping(
                    EventSourceArn=event_source_arn,
                    FunctionName=lambda_alias_arn,
                    BatchSize=rules.
                    get('batch_size', event_defaults[event_source]['batch_size']),
                    MaximumBatchingWindowInSeconds=rules.
                    get('batch_window', event_defaults[event_source]['batch_window']),
                    StartingPosition=starting_position,
                    StartingPositionTimestamp=datetime.utcfromtimestamp(
                        rules.
                        get('starting_position_timestamp', event_defaults[event_source]['starting_position_timestamp'])
                    ),
                    BisectBatchOnFunctionError=rules.
                    get('split_on_error', event_defaults[event_source]['split_on_error']),
                    MaximumRetryAttempts=rules.
                    get('max_retry', event_defaults[event_source]['max_retry']),
                    DestinationConfig=rules.
                    get('destination_config', event_defaults[event_source]['destination_config']),
                    MaximumRecordAgeInSeconds=rules.
                    get('max_record_age', event_defaults[event_source]['max_record_age']),
                    ParallelizationFactor=rules.
                    get('parallelization_factor', event_defaults[event_source]['parallelization_factor']))
            else:
                lambda_client.create_event_source_mapping(
                    EventSourceArn=event_source_arn,
                    FunctionName=lambda_alias_arn,
                    BatchSize=rules.
                    get('batch_size', event_defaults[event_source]['batch_size']),
                    MaximumBatchingWindowInSeconds=rules.
                    get('batch_window', event_defaults[event_source]['batch_window']),
                    StartingPosition=starting_position,
                    BisectBatchOnFunctionError=rules.
                    get('split_on_error', event_defaults[event_source]['split_on_error']),
                    MaximumRetryAttempts=rules.
                    get('max_retry', event_defaults[event_source]['max_retry']),
                    DestinationConfig=rules.
                    get('destination_config', event_defaults[event_source]['destination_config']),
                    MaximumRecordAgeInSeconds=rules.
                    get('max_record_age', event_defaults[event_source]['max_record_age']),
                    ParallelizationFactor=rules.
                    get('parallelization_factor', event_defaults[event_source]['parallelization_factor']))
        LOG.debug('{0} event trigger created'.format(event_defaults[event_source]['service_name']))

    LOG.info('Created {} event trigger on {} for {}'.format(event_defaults[event_source]['service_name'],
                                                            lambda_alias_arn, event_source_arn))



"""Verifies that instance_links are being retrieved properly from LINKS. Verifies that app_data.json.j2
contains the instance link information"""
from unittest import mock
from foremast.app import SpinnakerApp


@mock.patch('foremast.app.spinnaker_app.LINKS', new={'example1': 'https://example1.com'})
def test_default_instance_links():
    """Validate default instance_links are being populated properly."""
    pipeline_config = {
        "instance_links": {
            "example2": "https://example2.com",
        }
    }

    combined = {'example1': 'https://example1.com'}
    combined.update(pipeline_config['instance_links'])

    spinnaker_app = SpinnakerApp("aws", pipeline_config=pipeline_config)
    instance_links = spinnaker_app.retrieve_instance_links()

    assert instance_links == combined, "Instance Links are not being retrieved properly"


@mock.patch('foremast.app.spinnaker_app.LINKS', new={'example': 'example1', 'example': 'example2'})
def test_duplicate_instance_links():
    """Validate behavior when two keys are identical."""

    pipeline_config = {
        "instance_links": {}
    }

    duplicate = {'example': 'example2'}
    spinnaker_app = SpinnakerApp("aws", pipeline_config=pipeline_config)
    instance_links = spinnaker_app.retrieve_instance_links()

    assert instance_links == duplicate, "Instance links handing duplicates are wrong."



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Create Data Pipelines"""

import logging

import boto3
from awscli.customizations.datapipeline import translator

from ..exceptions import DataPipelineDefinitionError
from ..utils import get_details, get_properties

LOG = logging.getLogger(__name__)


class AWSDataPipeline:
    """Manipulate Data Pipeline."""

    def __init__(self, app=None, env=None, region='us-east-1', prop_path=None):
        """AWS Data Pipeline object.

        Args:
            app (str): Application name
            env (str): Environment/Account
            region (str): AWS Region
            prop_path (str): Path of environment property file
        """
        self.app_name = app
        self.env = env
        self.region = region
        self.properties = get_properties(prop_path, env=self.env, region=self.region)
        self.datapipeline_data = self.properties['datapipeline']
        generated = get_details(app=self.app_name)
        self.group = generated.data['project']

        session = boto3.Session(profile_name=self.env, region_name=self.region)
        self.client = session.client('datapipeline')
        self.pipeline_id = None

    def create_datapipeline(self):
        """Creates the data pipeline if it does not already exist

        Returns:
                dict: the response of the Boto3 command
        """

        tags = [{"key": "app_group", "value": self.group}, {"key": "app_name", "value": self.app_name}]
        response = self.client.create_pipeline(
            name=self.datapipeline_data.get('name', self.app_name),
            uniqueId=self.app_name,
            description=self.datapipeline_data['description'],
            tags=tags)
        self.pipeline_id = response.get('pipelineId')

        LOG.debug(response)
        LOG.info("Successfully configured Data Pipeline - %s", self.app_name)
        return response

    def set_pipeline_definition(self):
        """Translates the json definition and puts it on created pipeline

        Returns:
                dict: the response of the Boto3 command
        """

        if not self.pipeline_id:
            self.get_pipeline_id()

        json_def = self.datapipeline_data['json_definition']
        try:
            pipelineobjects = translator.definition_to_api_objects(json_def)
            parameterobjects = translator.definition_to_api_parameters(json_def)
            parametervalues = translator.definition_to_parameter_values(json_def)
        except translator.PipelineDefinitionError as error:
            LOG.warning(error)
            raise DataPipelineDefinitionError

        response = self.client.put_pipeline_definition(
            pipelineId=self.pipeline_id,
            pipelineObjects=pipelineobjects,
            parameterObjects=parameterobjects,
            parameterValues=parametervalues)
        LOG.debug(response)
        LOG.info("Successfully applied pipeline definition")
        return response

    def get_pipeline_id(self):
        """Finds the pipeline ID for configured pipeline"""

        all_pipelines = []
        paginiator = self.client.get_paginator('list_pipelines')
        for page in paginiator.paginate():
            all_pipelines.extend(page['pipelineIdList'])

        for pipeline in all_pipelines:
            if pipeline['name'] == self.datapipeline_data.get('name', self.app_name):
                self.pipeline_id = pipeline['id']
                LOG.info("Pipeline ID Found")
                return
        LOG.info("Pipeline ID Not Found for %s", self.app_name)

    def activate_pipeline(self):
        """Activates a deployed pipeline, useful for OnDemand pipelines"""
        self.client.activate_pipeline(pipelineId=self.pipeline_id)
        LOG.info("Activated Pipeline %s", self.pipeline_id)



import logging

LOG = logging.getLogger(__name__)


def exponential_backoff(attempt, exponent=2) -> float:
    """Exponential backoff function that returns a wait in seconds.  The attempt is multiplied exponentially
    using the exponent value (default 2).  For example: attempt 2 with an exponent of 4 would be 2^4=16.  This
    creates a backoff that gets exponentially longer for each attempt mitigating rate limiting and resource
    conflict errors.

    Use with retries annotation:

    Defaults:
    @retries(max_attempts=3, wait=exponential_backoff, exceptions=MyError)

    Overriding defaults:
    @retries(max_attempts=3, wait=lambda n: exponential_backoff(n, 4), exceptions=MyError)

    Args:
        attempt(int): Int representing the number of attempts so far
        exponent(float): Exponential modifier used to determine how long to wait
    """
    wait = exponent ** attempt
    LOG.debug("Backing off for {} seconds".format(wait))
    return wait



"""Test Provider Health Check setting."""
import copy

from foremast.pipeline.construct_pipeline_block import check_provider_healthcheck

TEST_SETTINGS = {'app': {'eureka_enabled': False}, 'asg': {'provider_healthcheck': {}}}


def test_provider_healthcheck():
    """Make sure default Provider Health Check works."""
    health_checks = check_provider_healthcheck(settings=TEST_SETTINGS)
    assert health_checks.providers == []
    assert health_checks.has_healthcheck == False


def test_setting_eureka_enabled():
    """When Eureka is enabled, default Helath Check should be used."""
    eureka_enabled_settings = copy.deepcopy(TEST_SETTINGS)
    eureka_enabled_settings['app']['eureka_enabled'] = True

    health_checks = check_provider_healthcheck(settings=eureka_enabled_settings)
    assert health_checks.providers == ['Discovery']
    assert health_checks.has_healthcheck is True


def test_defined_provider():
    """Provider defined in settings should be returned with capitalization."""
    provider_settings = copy.deepcopy(TEST_SETTINGS)
    provider_settings['asg']['provider_healthcheck']['cloud'] = True
    print(provider_settings)

    health_checks = check_provider_healthcheck(settings=provider_settings)
    assert len(health_checks.providers) == 1
    assert 'cloud' not in health_checks.providers
    assert 'Cloud' in health_checks.providers
    assert health_checks.has_healthcheck is True


def test_additional_provider_with_eureka():
    """Default Provider should be added to providers in settings."""
    eureka_enabled_with_provider_settings = copy.deepcopy(TEST_SETTINGS)
    eureka_enabled_with_provider_settings['app']['eureka_enabled'] = True
    eureka_enabled_with_provider_settings['asg']['provider_healthcheck']['cloud'] = True

    health_checks = check_provider_healthcheck(settings=eureka_enabled_with_provider_settings)
    assert len(health_checks.providers) == 2
    assert 'Discovery' in health_checks.providers
    assert 'cloud' not in health_checks.providers
    assert 'Cloud' in health_checks.providers
    assert health_checks.has_healthcheck is True



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Prepare the Application Configurations."""
import collections
import logging

from .. import consts
from ..exceptions import ForemastError
from ..utils import DeepChainMap, FileLookup

LOG = logging.getLogger(__name__)


def process_git_configs(git_short=''):
    """Retrieve _application.json_ files from GitLab.

    Args:
        git_short (str): Short Git representation of repository, e.g.
            forrest/core.

    Returns:
        collections.defaultdict: Configurations stored for each environment
        found.
    """
    LOG.info('Processing application.json files from GitLab "%s".', git_short)
    file_lookup = FileLookup(git_short=git_short)
    app_configs = process_configs(file_lookup,
                                  consts.RUNWAY_BASE_PATH + '/application-master-{env}.json',
                                  consts.RUNWAY_BASE_PATH + '/pipeline.json')
    commit_obj = file_lookup.project.commits.get('master')
    config_commit = commit_obj.attributes['id']
    LOG.info('Commit ID used: %s', config_commit)
    app_configs['pipeline']['config_commit'] = config_commit
    return app_configs


def process_runway_configs(runway_dir=''):
    """Read the _application.json_ files.

    Args:
        runway_dir (str): Name of runway directory with app.json files.

    Returns:
        collections.defaultdict: Configurations stored for each environment
        found.
    """
    LOG.info('Processing application.json files from local directory "%s".', runway_dir)
    file_lookup = FileLookup(runway_dir=runway_dir)
    app_configs = process_configs(file_lookup, 'application-master-{env}.json', 'pipeline.json')
    return app_configs


def process_configs(file_lookup, app_config_format, pipeline_config):
    """Processes the configs from lookup sources.

    Args:
        file_lookup (FileLookup): Source to look for file/config
        app_config_format (str): The format for application config files.
        pipeline_config (str): Name/path of the pipeline config

    Returns:
        dict: Retrieved application config
    """
    app_configs = collections.defaultdict(dict)
    # Load pipeline config first, determine cloud provider
    # Then load appropriate environment files
    try:
        app_configs['pipeline'] = file_lookup.json(filename=pipeline_config)
    except FileNotFoundError:
        LOG.warning('Unable to process pipeline.json. Using defaults.')
        app_configs['pipeline'] = {'env': ['stage', 'prod']}

    if 'type' in app_configs['pipeline']:
        pipeline_type = app_configs['pipeline']['type']
    else:
        pipeline_type = 'ec2'
    cloud_provider = get_cloud_for_pipeline_type(pipeline_type)
    environments = _get_env_names_for_cloud(cloud_provider)
    LOG.info("Using cloud provider '%s' for pipeline type '%s', supported environments: '%s'",
             cloud_provider, pipeline_type, environments)

    for env in environments:
        file_json = app_config_format.format(env=env)
        try:
            env_config = file_lookup.json(filename=file_json)
            app_configs[env] = apply_region_configs(env_config)
        except FileNotFoundError:
            LOG.critical('Application configuration not available for %s.', env)
            continue

    LOG.debug('Application configs:\n%s', app_configs)
    return app_configs


def apply_region_configs(env_config):
    """Override default env configs with region specific configs and nest
    all values under a region

    Args:
        env_config (dict): The environment specific config.

    Return:
        dict: Newly updated dictionary with region overrides applied.
    """
    new_config = env_config.copy()
    for region in env_config.get('regions', consts.REGIONS):
        if isinstance(env_config.get('regions'), dict):
            region_specific_config = env_config['regions'][region]
            new_config[region] = dict(DeepChainMap(region_specific_config, env_config))
        else:
            new_config[region] = env_config.copy()
    LOG.debug('Region Specific Config:\n%s', new_config)
    return new_config


def get_cloud_for_pipeline_type(pipeline_type):
    """Maps a pipeline type to the corresponding cloud provider

        Args:
            pipeline_type (str): The pipeline type (e.g. cloudfunction, lambda, ec2)

        Return:
            str: The corresponding cloud provider (e.g. aws, gcp)
        """
    if pipeline_type in consts.GCP_TYPES:
        return "gcp"
    elif pipeline_type in consts.AWS_TYPES:
        return "aws"
    else:
        error_message = ("pipeline.type of '{0}' is not supported. "
                         "If this is a manual pipeline it is required you specify the "
                         "pipeline type in AWS_MANUAL_TYPES or GCP_MANUAL_TYPES"
                         ).format(pipeline_type)
        raise ForemastError(error_message)


def _get_env_names_for_cloud(cloud_name):
    """Returns the list of env names for the given cloud

    Args:
        cloud_name, Str: Name of cloud provider.  'aws' and 'gcp' are only supported options.

    Return:
        set: Environment names for the given cloud (e.g. stage, prod) and corresponding configs
    """
    if cloud_name == "aws":
        return consts.ENVS
    elif cloud_name == "gcp":
        # GCP env config is an nested dictionary of env names with config as value
        # only return the names as that is all that is needed, and the AWS config is names only
        return set(consts.GCP_ENVS.keys())
    else:
        raise ValueError("Unknown cloud given while loading cloud environments. Only gcp and aws are acceptable.  "
                         "Received '%s'", cloud_name)



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Check Pipeline name to match format."""
import logging
import uuid

from ..utils.gate import gate_request

LOG = logging.getLogger(__name__)

# Random UUID namespace for use when generating pipeline Ids locally
# This should never change!
_foremast_uuid_namespace = uuid.UUID("098f0a30-180e-4aa4-b9bd-72d63fda3c18")


def check_managed_pipeline(name='', app_name=''):
    """Check a Pipeline name is a managed format **app_name [region]**.

    Args:
        name (str): Name of Pipeline to check.
        app_name (str): Name of Application to find in Pipeline name.

    Returns:
        str: Region name from managed Pipeline name.

    Raises:
        ValueError: Pipeline is not managed.

    """
    *pipeline_name_prefix, bracket_region = name.split()
    region = bracket_region.strip('[]')

    not_managed_message = '"{0}" is not managed.'.format(name)

    if 'onetime' in region:
        LOG.info('"%s" is a onetime, marked for cleaning.', name)
        return region

    if not all([bracket_region.startswith('['), bracket_region.endswith(']')]):
        LOG.debug('"%s" does not end with "[region]".', name)
        raise ValueError(not_managed_message)

    if len(pipeline_name_prefix) != 1:
        LOG.debug('"%s" does not only have one word before [region].', name)
        raise ValueError(not_managed_message)

    if app_name not in pipeline_name_prefix:
        LOG.debug('"%s" does not use "%s" before [region].', name, app_name)
        raise ValueError(not_managed_message)

    return region


def get_all_pipelines(app=''):
    """Get a list of all the Pipelines in _app_.

    Args:
        app (str): Name of Spinnaker Application.

    Returns:
        requests.models.Response: Response from Gate containing Pipelines.

    """
    uri = '/applications/{app}/pipelineConfigs'.format(app=app)
    response = gate_request(uri=uri)

    assert response.ok, 'Could not retrieve Pipelines for {0}.'.format(app)

    pipelines = response.json()
    LOG.debug('Pipelines:\n%s', pipelines)

    return pipelines


def get_pipeline_id(app='', name=''):
    """Get the ID for Pipeline _name_.

    Args:
        app (str): Name of Spinnaker Application to search.
        name (str): Name of Pipeline to get ID for.

    Returns:
        str: ID of specified Pipeline.
        None: Pipeline or Spinnaker Application not found.
    """
    return_id = None

    pipelines = get_all_pipelines(app=app)

    for pipeline in pipelines:
        LOG.debug('ID of %(name)s: %(id)s', pipeline)

        if pipeline['name'] == name:
            return_id = pipeline['id']
            LOG.info('Pipeline %s found, ID: %s', name, return_id)
            break

    return return_id


def normalize_pipeline_name(name=''):
    """Translate unsafe characters to underscores."""
    normalized_name = name
    for bad in '\\/?%#':
        normalized_name = normalized_name.replace(bad, '_')
    return normalized_name


def generate_predictable_pipeline_id(application_name, pipeline_name):
    """Create a predictable pipeline ID (UUID) using an application name and pipeline name.
    It will always produce the same UUID when given the same arguments

    Args:
        application_name (str): Name of Spinnaker Application
        pipeline_name (str): Name of the pipeline

    Returns:
        UUID: UUID generated using the seed arguments
    """
    seed = application_name + pipeline_name
    pipeline_uuid = uuid.uuid5(_foremast_uuid_namespace, seed)
    LOG.debug("Generating pipeline id '%s' using namespace '%s' and seed '%s'",
              pipeline_name, _foremast_uuid_namespace, seed)
    return pipeline_uuid



"""S3 Tag functions."""


def generated_tag_data(tags):
    """Convert :obj:`dict` to S3 Tag list.

    Args:
        tags (dict): Dictonary of tag key and tag value passed.

    Returns:
        list: List of dictionaries.

    """
    generated_tags = []
    for key, value in tags.items():
        generated_tags.append({
            'Key': key,
            'Value': value,
        })
    return generated_tags



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Foremast and Spinnaker related custom exceptions."""
import logging
import pprint

from .consts import GOOD_STATUSES, SKIP_STATUSES

LOG = logging.getLogger(__name__)


class ForemastError(Exception):
    """Foremast related error."""


class ForemastTemplateNotFound(Exception):
    """Foremast Template was not found."""


class ForemastConfigurationFileError(ForemastError):
    """Foremast configuration file misconfigured."""


class GitLabApiError(ForemastError):
    """GitLab API did not return a good status."""


class GoogleIAPTokenError(ForemastError):
    """Google IAP Token Request did not return a good status."""


class GoogleIAPError(ForemastError):
    """Google IAP did not return a good status."""


class GoogleInfrastructureError(ForemastError):
    """Error while executing Google Cloud Platform Infrastructure steps"""


class SpinnakerError(ForemastError):
    """Spinnaker related error."""


class SpinnakerAppNotFound(SpinnakerError):
    """Spinnaker app not found error."""


class SpinnakerApplicationListError(SpinnakerError):
    """Spinnaker application list error."""


class SpinnakerDnsCreationFailed(SpinnakerError):
    """Spinnaker DNS creation error."""


class SpinnakerElbNotFound(SpinnakerError):
    """Spinnaker Elb not found."""


class SpinnakerTimeout(SpinnakerError):
    """Spinnaker Timeout error."""


class SpinnakerVPCNotFound(SpinnakerError):
    """Spinnaker did not find a VPC."""


class SpinnakerVPCIDNotFound(SpinnakerError):
    """Spinnaker did not find the VPC ID."""


class SpinnakerTaskError(SpinnakerError):
    """Spinnaker Task did not finish properly."""

    def __init__(self, task_state):
        errors = []

        skip_statuses = GOOD_STATUSES.union(SKIP_STATUSES)

        for stage in task_state['execution']['stages']:
            LOG.debug('Stage:\n%s', pprint.pformat(stage))

            if stage.get('status') in skip_statuses:
                continue

            context = stage['context']

            try:
                errors.extend(context['exception']['details']['errors'])
            except KeyError:
                for task in context['kato.tasks']:
                    errors.append(task['exception']['message'])

        super().__init__(*errors)


class SpinnakerTaskInconclusiveError(SpinnakerTaskError):
    """Spinnaker Task state failed to reach terminal state in allowed time."""

    def __init__(self, message):
        mock_failure_stage = {
            'context': {
                'exception': {
                    'details': {
                        'errors': [message],
                    },
                },
            },
        }
        spinnaker_task_state = {
            'execution': {
                'stages': [mock_failure_stage],
            },
        }

        super().__init__(spinnaker_task_state)


class SpinnakerPipelineCreationFailed(SpinnakerError):
    """Could not create Spinnaker Pipeline."""


class SpinnakerPipelineDeletionFailed(SpinnakerError):
    """Could not delete Spinnaker Pipeline."""


class SpinnakerSecurityGroupCreationFailed(SpinnakerError):
    """Could not create Security Group."""


class SpinnakerSecurityGroupError(SpinnakerError):
    """Could not create Security Group."""


class SpinnakerSubnetError(SpinnakerError):
    """Unavailable environment or region."""

    def __init__(self, env='', region=''):
        error = '{0} is not available for {1}'.format(region, env)
        super().__init__(error)


class InvalidEventConfiguration(ForemastError):
    """Invalid AWS Lambda event configuration."""


class DynamoDBTableNotFound(ForemastError):
    """DynamoDB Table was not found."""


class DynamoDBStreamNotFound(ForemastError):
    """DynamoDB Table Streams was not found."""


class SNSTopicNotFound(ForemastError):
    """SNS Topic was not found."""


class SNSSubscriptionDoesNotExist(ForemastError):
    """SNS Subscriptions does not exist."""


class LambdaFunctionDoesNotExist(ForemastError):
    """Lambda function was not found."""


class LambdaAliasDoesNotExist(ForemastError):
    """Lambda function was not found."""


class RequiredKeyNotFound(ForemastError):
    """Required key in json config not found."""


class PrimaryDNSRecordNotFound(ForemastError):
    """Required Primary DNS record does not exist."""


class S3ArtifactNotFound(ForemastError):
    """Could not find Artifact to upload to S3."""


class S3SharedBucketNotFound(ForemastError):
    """Shared S3 Bucket does not exist."""


class DataPipelineDefinitionError(ForemastError):
    """Error Creating Data Pipeline."""


class StepFunctionDefinitionError(ForemastError):
    """Error Creating Step Function."""


class CloudFunctionDeployError(ForemastError):
    """Error while deploying a Cloud Function"""


class CloudFunctionOperationIncompleteError(CloudFunctionDeployError):
    """Error waiting on GCP Cloud Function operation to complete"""


class CloudFunctionOperationFailedError(CloudFunctionDeployError):
    """Cloud Function operation completed with errors"""

    def __init__(self, operation_error: dict, message: str):
        self.operation_error = operation_error
        self.message = message
        super().__init__(self.message)



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Destroy any ELB Resources."""

from ...utils import get_template, get_vpc_id, wait_for_task


def destroy_elb(app='', env='dev', region='us-east-1', **_):
    """Destroy ELB Resources.

    Args:
        app (str): Spinnaker Application name.
        env (str): Deployment environment.
        region (str): AWS region.

    Returns:
        True upon successful completion.
    """
    task_json = get_template(
        template_file='destroy/destroy_elb.json.j2',
        app=app,
        env=env,
        region=region,
        vpc=get_vpc_id(account=env, region=region))

    wait_for_task(task_json)

    return True



#   Foremast - Pipeline Tooling
#
#   Copyright 2019 Redbox Automated Retail, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Set of utility functions for Spinnaker's kayenta canary service"""
from ..exceptions import SpinnakerPipelineCreationFailed
from ..utils.gate import gate_request


def get_canary_id(name, application=None):
    """Finds a canary config ID matching the name passed.
    Assumes the canary name is unique and the first match wins.
    Args:
        application (str): Name of Spinnaker Application to search (optional)
        name (str): Name of CanaryConfig to get the ID for

    Returns:
        str: ID of specified CanaryConfig.
        None: CanaryConfig not found
    """
    uri = '/v2/canaryConfig'
    canary_response = gate_request(uri=uri)

    if not canary_response.ok:
        raise SpinnakerPipelineCreationFailed('Could not resolve canary config id for {0}: {1}'
                                              .format(name, canary_response.json()))

    canary_options = canary_response.json()
    names = []
    for config in canary_options:
        names.append(config['name'])
        # If this canary name matches and they did not specificy the owning application
        if config['name'] == name and application is None:
            return config['id']
        # If this canary name matches and the application is listed in the canary config's owners array
        if config['name'] == name and application in config['applications']:
            return config['id']

    raise SpinnakerPipelineCreationFailed(
        'Could not resolve canary config id for: {0}.  Options are: {1}'.format(name, names))



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Renumerate the Pipeline Stages."""
import logging

LOG = logging.getLogger(__name__)


def renumerate_stages(pipeline):
    """Renumber Pipeline Stage reference IDs to account for dependencies.

    stage order is defined in the templates. The ``refId`` field dictates
    if a stage should be mainline or parallel to other stages.

        * ``master`` - A mainline required stage. Other stages depend on it
        * ``branch`` - A stage that should be ran in parallel to master stages.
        * ``merge`` - A stage thatis parallel but other stages still depend on it.

    Args:
        pipeline (dict): Completed Pipeline ready for renumeration.

    Returns:
        dict: Pipeline ready to be sent to Spinnaker.
    """
    stages = pipeline['stages']

    main_index = 0
    branch_index = 0
    previous_refid = ''

    for stage in stages:
        current_refid = stage['refId'].lower()
        if current_refid == 'master':
            if main_index == 0:
                stage['requisiteStageRefIds'] = []
            else:
                stage['requisiteStageRefIds'] = [str(main_index)]
            main_index += 1
            stage['refId'] = str(main_index)
        elif current_refid == 'branch':
            # increments a branch_index to account for multiple parrallel stages
            if previous_refid == 'branch':
                branch_index += 1
            else:
                branch_index = 0
            stage['refId'] = str((main_index * 100) + branch_index)
            stage['requisiteStageRefIds'] = [str(main_index)]
        elif current_refid == 'merge':
            # TODO: Added logic to handle merge stages.
            pass

        previous_refid = current_refid
        LOG.debug('step=%(name)s\trefId=%(refId)s\t' 'requisiteStageRefIds=%(requisiteStageRefIds)s', stage)

    return pipeline



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Archaius functions for deployment."""
import logging

import boto3

from ..utils import get_details

LOG = logging.getLogger(__name__)


def init_properties(env='dev', app='unnecessary', **_):
    """Make sure _application.properties_ file exists in S3.

    For Applications with Archaius support, there needs to be a file where the
    cloud environment variable points to.

    Args:
        env (str): Deployment environment/account, i.e. dev, stage, prod.
        app (str): GitLab Project name.

    Returns:
        True when application.properties was found.
        False when application.properties needed to be created.
    """
    aws_env = boto3.session.Session(profile_name=env)
    s3client = aws_env.resource('s3')

    generated = get_details(app=app, env=env)
    archaius = generated.archaius()

    archaius_file = ('{path}/application.properties').format(path=archaius['path'])

    try:
        s3client.Object(archaius['bucket'], archaius_file).get()
        LOG.info('Found: %(bucket)s/%(file)s', {'bucket': archaius['bucket'], 'file': archaius_file})
        return True
    except boto3.exceptions.botocore.client.ClientError:
        s3client.Object(archaius['bucket'], archaius_file).put()
        LOG.info('Created: %(bucket)s/%(file)s', {'bucket': archaius['bucket'], 'file': archaius_file})
        return False
